import pytest
import numpy as np

from xeofs.preprocessing.sanitizer import Sanitizer
from xeofs.utils.data_types import DataArray
from ..conftest import generate_synthetic_dataarray
from ..utilities import (
    data_is_dask,
    assert_expected_dims,
    assert_expected_coords,
)

# =============================================================================
# VALID TEST CASES
# =============================================================================
N_SAMPLE_DIMS = [1]
N_FEATURE_DIMS = [1]
INDEX_POLICY = ["index"]
NAN_POLICY = ["no_nan", "fulldim"]
DASK_POLICY = ["no_dask", "dask"]
SEED = [0]

VALID_TEST_DATA = [
    (ns, nf, index, nan, dask)
    for ns in N_SAMPLE_DIMS
    for nf in N_FEATURE_DIMS
    for index in INDEX_POLICY
    for nan in NAN_POLICY
    for dask in DASK_POLICY
]

FULL_NAN = [
    (ns, nf, index, nan, dask)
    for ns in N_SAMPLE_DIMS
    for nf in N_FEATURE_DIMS
    for index in INDEX_POLICY
    for nan in ["fulldim"]
    for dask in DASK_POLICY
]

ISOLATED_NAN = [
    (ns, nf, index, nan, dask)
    for ns in N_SAMPLE_DIMS
    for nf in N_FEATURE_DIMS
    for index in INDEX_POLICY
    for nan in ["isolated"]
    for dask in DASK_POLICY
]


# TESTS
# =============================================================================
@pytest.mark.parametrize(
    "sample_name, feature_name, data_params",
    [
        ("sample", "feature", (1, 1)),
        ("another_sample", "another_feature", (1, 1)),
    ],
)
def test_fit_valid_dimension_names(sample_name, feature_name, data_params):
    data = generate_synthetic_dataarray(*data_params)
    data = data.rename({"sample0": sample_name, "feature0": feature_name})

    sanitizer = Sanitizer(sample_name=sample_name, feature_name=feature_name)
    sanitizer.fit(data)
    data_clean = sanitizer.transform(data)
    reconstructed_data = sanitizer.inverse_transform_data(data_clean)

    assert data_clean.ndim == 2
    assert set(data_clean.dims) == set((sample_name, feature_name))
    assert set(reconstructed_data.dims) == set(data.dims)


@pytest.mark.parametrize(
    "sample_name, feature_name, data_params",
    [
        ("sample1", "feature", (1, 1)),
        ("sample", "feature1", (1, 1)),
        ("sample1", "feature1", (1, 1)),
    ],
)
def test_fit_invalid_dimension_names(sample_name, feature_name, data_params):
    data = generate_synthetic_dataarray(*data_params)

    sanitizer = Sanitizer(sample_name=sample_name, feature_name=feature_name)

    with pytest.raises(ValueError):
        sanitizer.fit(data)


@pytest.mark.parametrize(
    "synthetic_dataarray",
    VALID_TEST_DATA,
    indirect=["synthetic_dataarray"],
)
def test_transform(synthetic_dataarray):
    data = synthetic_dataarray
    data = data.rename({"sample0": "sample", "feature0": "feature"})

    sanitizer = Sanitizer()
    sanitizer.fit(data)
    transformed_data = sanitizer.transform(data)
    transformed_data2 = sanitizer.transform(data)

    is_dask_before = data_is_dask(data)
    is_dask_after = data_is_dask(transformed_data)

    assert transformed_data.notnull().all()
    assert isinstance(transformed_data, DataArray)
    assert transformed_data.ndim == 2
    assert transformed_data.dims == data.dims
    assert is_dask_before == is_dask_after
    assert transformed_data.identical(transformed_data2)


@pytest.mark.parametrize(
    "synthetic_dataarray",
    VALID_TEST_DATA,
    indirect=["synthetic_dataarray"],
)
def test_transform_invalid(synthetic_dataarray):
    data = synthetic_dataarray
    data = data.rename({"sample0": "sample", "feature0": "feature"})

    sanitizer = Sanitizer()
    sanitizer.fit(data)
    with pytest.raises(ValueError):
        sanitizer.transform(data.isel(feature0=slice(0, 2)))


@pytest.mark.parametrize(
    "synthetic_dataarray",
    VALID_TEST_DATA,
    indirect=["synthetic_dataarray"],
)
def test_fit_transform(synthetic_dataarray):
    data = synthetic_dataarray
    data = data.rename({"sample0": "sample", "feature0": "feature"})

    sanitizer = Sanitizer()
    transformed_data = sanitizer.fit_transform(data)

    is_dask_before = data_is_dask(data)
    is_dask_after = data_is_dask(transformed_data)

    assert isinstance(transformed_data, DataArray)
    assert transformed_data.notnull().all()
    assert transformed_data.ndim == 2
    assert transformed_data.dims == data.dims
    assert is_dask_before == is_dask_after


@pytest.mark.parametrize(
    "synthetic_dataarray",
    VALID_TEST_DATA,
    indirect=["synthetic_dataarray"],
)
def test_invserse_transform_data(synthetic_dataarray):
    data = synthetic_dataarray
    data = data.rename({"sample0": "sample", "feature0": "feature"})

    sanitizer = Sanitizer()
    sanitizer.fit(data)
    cleaned_data = sanitizer.transform(data)
    uncleaned_data = sanitizer.inverse_transform_data(cleaned_data)

    is_dask_before = data_is_dask(data)
    is_dask_after = data_is_dask(uncleaned_data)

    # inverse transform is only identical if nan_policy={"no_nan", "fulldim"}
    # in case of "isolated" the inverse transform will set the entire feature column
    # to NaNs, which is not identical to the original data
    # assert data.identical(uncleaned_data)

    # inverse transform should not change dask-ness
    assert is_dask_before == is_dask_after


@pytest.mark.parametrize(
    "synthetic_dataarray",
    VALID_TEST_DATA,
    indirect=["synthetic_dataarray"],
)
def test_invserse_transform_components(synthetic_dataarray):
    data: DataArray = synthetic_dataarray
    data = data.rename({"sample0": "sample", "feature0": "feature"})

    sanitizer = Sanitizer()
    sanitizer.fit(data)

    stacked_data = sanitizer.transform(data)
    components = stacked_data.rename({"sample": "mode"})
    unstacked_data = sanitizer.inverse_transform_components(components)

    is_dask_before = data_is_dask(data)
    is_dask_after = data_is_dask(unstacked_data)

    # Unstacked components has correct feature dimensions
    assert_expected_dims(data, unstacked_data, policy="feature")
    # Unstacked data has coordinates of original data
    assert_expected_coords(data, unstacked_data, policy="feature")
    # inverse transform should not change dask-ness
    assert is_dask_before == is_dask_after


@pytest.mark.parametrize(
    "synthetic_dataarray",
    VALID_TEST_DATA,
    indirect=["synthetic_dataarray"],
)
def test_invserse_transform_scores(synthetic_dataarray):
    data: DataArray = synthetic_dataarray
    data = data.rename({"sample0": "sample", "feature0": "feature"})

    sanitizer = Sanitizer()
    sanitizer.fit(data)

    stacked_data = sanitizer.transform(data)
    components = stacked_data.rename({"feature": "mode"})
    unstacked_data = sanitizer.inverse_transform_scores(components)

    is_dask_before = data_is_dask(data)
    is_dask_after = data_is_dask(unstacked_data)

    # Unstacked components has correct feature dimensions
    assert_expected_dims(data, unstacked_data, policy="sample")
    # Unstacked data has coordinates of original data
    assert_expected_coords(data, unstacked_data, policy="sample")
    # inverse transform should not change dask-ness
    assert is_dask_before == is_dask_after


@pytest.mark.parametrize(
    "synthetic_dataarray",
    ISOLATED_NAN,
    indirect=["synthetic_dataarray"],
)
def test_isolated_nans(synthetic_dataarray):
    data = synthetic_dataarray
    data = data.rename({"sample0": "sample", "feature0": "feature"})

    sanitizer = Sanitizer()
    sanitizer.fit(data)

    # By default want the sanitizer to raise for isolated NaNs
    with pytest.raises(ValueError):
        sanitizer.transform(data)

    # But allow this check to be disabled
    sanitizer.check_nans = False
    sanitizer.transform(data)


@pytest.mark.parametrize(
    "synthetic_dataarray",
    FULL_NAN,
    indirect=["synthetic_dataarray"],
)
def test_feature_nan_transform(synthetic_dataarray):
    data = synthetic_dataarray
    data = data.rename({"sample0": "sample", "feature0": "feature"})

    sanitizer = Sanitizer()
    sanitizer.fit(data)
    sanitizer.transform(data)

    # Pass through new data with NaNs in a different location
    data.loc[{"feature": 1}] = np.nan
    # By default want the sanitizer to raise on transform for new NaN features
    with pytest.raises(ValueError):
        sanitizer.transform(data)

    # But allow this check to be disabled
    sanitizer.check_nans = False
    sanitizer.transform(data)



VALID_LATITUDE_NAMES = [
    "latitude",
    "lats",
    "lat",
    "Latitude",
    "Lats",
    "Lat",
    "LATITUDE",
    "LATS",
    "LAT",
]

VALID_LONGITUDE_NAMES = ["lon", "lons", "longitude", "longitudes"]
VALID_CARTESIAN_X_NAMES = ["x", "x_coord"]
VALID_CARTESIAN_Y_NAMES = ["y", "y_coord"]

VALID_KERNELS = ["bisquare", "gaussian", "exponential"]
VALID_METRICS = ["euclidean", "haversine"]


MULTIPLE_TESTS = [
    "bonferroni",
    "sidak",
    "holm-sidak",
    "holm",
    "simes-hochberg",
    "hommel",
    "fdr_bh",
    "fdr_by",
    "fdr_tsbh",
    "fdr_tsbky",
]


AVG_EARTH_RADIUS = 6371.0  # in km



import warnings

from .cca import CCA

__all__ = ["CCA"]


DEPRECATED_NAMES = [
    # ("OldClass", "NewClass"),
]


def __dir__():
    return sorted(__all__ + [names[0] for names in DEPRECATED_NAMES])


def __getattr__(name):
    for old_name, new_name in DEPRECATED_NAMES:
        if name == old_name:
            msg = (
                f"Class '{old_name}' is deprecated and will be renamed to '{new_name}' in the next major release. "
                f"In that release, '{old_name}' will refer to a different class. "
                f"Please switch to '{new_name}' to maintain compatibility."
            )
            warnings.warn(msg, DeprecationWarning, stacklevel=2)
            return globals()[new_name]
    raise AttributeError(f"module {__name__} has no attribute {name}")



import numpy as np
import pytest
import xarray as xr

from xeofs.single import EOF

from ...utilities import skip_if_missing_engine


def test_init():
    """Tests the initialization of the EOF class"""
    eof = EOF(n_modes=5, standardize=True, use_coslat=True)

    # Assert preprocessor has been initialized
    assert hasattr(eof, "_params")
    assert hasattr(eof, "preprocessor")


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_fit(dim, mock_data_array):
    """Tests the fit method of the EOF class"""

    eof = EOF()
    eof.fit(mock_data_array, dim)

    # Assert the required attributes have been set
    assert hasattr(eof, "preprocessor")
    assert hasattr(eof, "data")


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_singular_values(dim, mock_data_array):
    """Tests the singular_values method of the EOF class"""

    eof = EOF()
    eof.fit(mock_data_array, dim)

    # Test singular_values method
    singular_values = eof.singular_values()
    assert isinstance(singular_values, xr.DataArray)


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_explained_variance(dim, mock_data_array):
    """Tests the explained_variance method of the EOF class"""
    eof = EOF()
    eof.fit(mock_data_array, dim)

    # Test explained_variance method
    explained_variance = eof.explained_variance()
    assert isinstance(explained_variance, xr.DataArray)
    # Explained variance must be positive
    assert (explained_variance > 0).all()


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_explained_variance_ratio(dim, mock_data_array):
    """Tests the explained_variance_ratio method of the EOF class"""
    eof = EOF()
    eof.fit(mock_data_array, dim)

    # Test explained_variance_ratio method
    explained_variance_ratio = eof.explained_variance_ratio()
    assert isinstance(explained_variance_ratio, xr.DataArray)
    # Explained variance ratio must be positive
    assert (
        explained_variance_ratio > 0
    ).all(), "The explained variance ratio must be positive"
    # The sum of the explained variance ratio must be <= 1
    assert (
        explained_variance_ratio.sum() <= 1 + 1e-5
    ), "The sum of the explained variance ratio must be <= 1"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_isolated_nans(dim, mock_data_array_isolated_nans):
    """Tests the components method of the EOF class"""
    eof = EOF()
    with pytest.raises(ValueError):
        eof.fit(mock_data_array_isolated_nans, dim)


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components(dim, mock_data_array):
    """Tests the components method of the EOF class"""
    eof = EOF()
    eof.fit(mock_data_array, dim)

    # Test components method
    components = eof.components()
    feature_dims = tuple(set(mock_data_array.dims) - set(dim))
    assert isinstance(components, xr.DataArray), "Components is not a DataArray"
    assert set(components.dims) == set(
        ("mode",) + feature_dims
    ), "Components does not have the right feature dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components_fulldim_nans(dim, mock_data_array_full_dimensional_nans):
    """Tests the components method of the EOF class"""
    eof = EOF()
    eof.fit(mock_data_array_full_dimensional_nans, dim)

    # Test components method
    components = eof.components()
    feature_dims = tuple(set(mock_data_array_full_dimensional_nans.dims) - set(dim))
    assert isinstance(components, xr.DataArray), "Components is not a DataArray"
    assert set(components.dims) == set(
        ("mode",) + feature_dims
    ), "Components does not have the right feature dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components_boundary_nans(dim, mock_data_array_boundary_nans):
    """Tests the components method of the EOF class"""
    eof = EOF()
    eof.fit(mock_data_array_boundary_nans, dim)

    # Test components method
    components = eof.components()
    feature_dims = tuple(set(mock_data_array_boundary_nans.dims) - set(dim))
    assert isinstance(components, xr.DataArray), "Components is not a DataArray"
    assert set(components.dims) == set(
        ("mode",) + feature_dims
    ), "Components does not have the right feature dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components_dataset(dim, mock_dataset):
    """Tests the components method of the EOF class"""
    eof = EOF()
    eof.fit(mock_dataset, dim)

    # Test components method
    components = eof.components()
    feature_dims = tuple(set(mock_dataset.dims) - set(dim))
    assert isinstance(components, xr.Dataset), "Components is not a Dataset"
    assert set(components.data_vars) == set(
        mock_dataset.data_vars
    ), "Components does not have the same data variables as the input Dataset"
    assert set(components.dims) == set(
        ("mode",) + feature_dims
    ), "Components does not have the right feature dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components_dataarray_list(dim, mock_data_array_list):
    """Tests the components method of the EOF class"""
    eof = EOF()
    eof.fit(mock_data_array_list, dim)

    # Test components method
    components = eof.components()
    feature_dims = [tuple(set(data.dims) - set(dim)) for data in mock_data_array_list]
    assert isinstance(components, list), "Components is not a list"
    assert len(components) == len(
        mock_data_array_list
    ), "Components does not have the same length as the input list"
    assert isinstance(
        components[0], xr.DataArray
    ), "Components is not a list of DataArrays"
    for comp, feat_dims in zip(components, feature_dims):
        assert set(comp.dims) == set(
            ("mode",) + feat_dims
        ), "Components does not have the right feature dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores(dim, mock_data_array):
    """Tests the scores method of the EOF class"""
    eof = EOF()
    eof.fit(mock_data_array, dim)

    # Test scores method
    scores = eof.scores()
    assert isinstance(scores, xr.DataArray), "Scores is not a DataArray"
    assert set(scores.dims) == set(
        (dim + ("mode",))
    ), "Scores does not have the right dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores_fulldim_nans(dim, mock_data_array_full_dimensional_nans):
    """Tests the scores method of the EOF class"""
    eof = EOF()
    eof.fit(mock_data_array_full_dimensional_nans, dim)

    # Test scores method
    scores = eof.scores()
    assert isinstance(scores, xr.DataArray), "Scores is not a DataArray"
    assert set(scores.dims) == set(
        (dim + ("mode",))
    ), "Scores does not have the right dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores_boundary_nans(dim, mock_data_array_boundary_nans):
    """Tests the scores method of the EOF class"""
    eof = EOF()
    eof.fit(mock_data_array_boundary_nans, dim)

    # Test scores method
    scores = eof.scores()
    assert isinstance(scores, xr.DataArray), "Scores is not a DataArray"
    assert set(scores.dims) == set(
        (dim + ("mode",))
    ), "Scores does not have the right dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores_dataset(dim, mock_dataset):
    """Tests the scores method of the EOF class"""
    eof = EOF()
    eof.fit(mock_dataset, dim)

    # Test scores method
    scores = eof.scores()
    assert isinstance(scores, xr.DataArray)
    assert set(scores.dims) == set(
        (dim + ("mode",))
    ), "Scores does not have the right dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores_dataarray_list(dim, mock_data_array_list):
    """Tests the scores method of the EOF class"""
    eof = EOF()
    eof.fit(mock_data_array_list, dim)

    # Test scores method
    scores = eof.scores()
    assert isinstance(scores, xr.DataArray)
    assert set(scores.dims) == set(
        (dim + ("mode",))
    ), "Scores does not have the right dimensions"


def test_get_params():
    """Tests the get_params method of the EOF class"""
    eof = EOF(n_modes=5, standardize=True, use_coslat=True)

    # Test get_params method
    params = eof.get_params()
    assert isinstance(params, dict)
    assert params.get("n_modes") == 5
    assert params.get("standardize") is True
    assert params.get("use_coslat") is True
    assert params.get("solver") == "auto"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_transform(dim, mock_data_array):
    """Test projecting new unseen data onto the components (EOFs/eigenvectors)"""
    data = mock_data_array.isel({dim[0]: slice(1, None)})
    new_data = mock_data_array.isel({dim[0]: slice(0, 1)})

    # Create a xarray DataArray with random data
    model = EOF(n_modes=2, solver="full")
    model.fit(data, dim)
    scores = model.scores()

    # Project data onto the components
    projections = model.transform(data)

    # Check that the projection has the right dimensions
    assert projections.dims == scores.dims, "Projection has wrong dimensions"  # type: ignore

    # Check that the projection has the right data type
    assert isinstance(projections, xr.DataArray), "Projection is not a DataArray"

    # Check that the projection has the right name
    assert projections.name == "scores", "Projection has wrong name: {}".format(
        projections.name
    )

    # Check that the projection's data is the same as the scores
    np.testing.assert_allclose(
        scores.sel(mode=slice(1, 3)), projections.sel(mode=slice(1, 3)), rtol=1e-3
    )

    # Project unseen data onto the components
    new_projections = model.transform(new_data)

    # Check that the projection has the right dimensions
    assert new_projections.dims == scores.dims, "Projection has wrong dimensions"  # type: ignore

    # Check that the projection has the right data type
    assert isinstance(new_projections, xr.DataArray), "Projection is not a DataArray"

    # Check that the projection has the right name
    assert new_projections.name == "scores", "Projection has wrong name: {}".format(
        new_projections.name
    )

    # Ensure that the new projections are not NaNs
    assert np.all(new_projections.notnull().values), "New projections contain NaNs"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_transform_nan_feature(dim, mock_data_array):
    """Test projecting new unseen data onto the components (EOFs/eigenvectors)"""
    data = mock_data_array.isel()

    # Create a xarray DataArray with random data
    model = EOF(n_modes=2, solver="full")
    model.fit(data, dim)

    # Set a new feature to NaN and attempt to project data onto the components
    feature_dims = list(set(data.dims) - set(dim))
    data_missing = data.copy()
    data_missing.loc[{feature_dims[0]: data[feature_dims[0]][0].values}] = np.nan

    # with nan checking, transform should fail if any new features are NaN
    with pytest.raises(ValueError):
        model.transform(data_missing)

    # without nan checking, transform will succeed but be all nan
    model = EOF(n_modes=2, solver="full", check_nans=False)
    model.fit(data, dim)

    data_transformed = model.transform(data_missing)
    assert data_transformed.isnull().all()


@pytest.mark.parametrize(
    "dim, normalized",
    [
        (("time",), True),
        (("lat", "lon"), True),
        (("lon", "lat"), True),
        (("time",), False),
        (("lat", "lon"), False),
        (("lon", "lat"), False),
    ],
)
def test_inverse_transform(dim, mock_data_array, normalized):
    """Test inverse_transform method in EOF class."""

    # instantiate the EOF class with necessary parameters
    eof = EOF(n_modes=20, standardize=True)

    # fit the EOF model
    eof.fit(mock_data_array, dim=dim)
    scores = eof.scores(normalized=normalized)

    # Test with single mode
    scores_selection = scores.sel(mode=1)
    X_rec_1 = eof.inverse_transform(scores_selection)
    assert isinstance(X_rec_1, xr.DataArray)

    # Test with single mode as list
    scores_selection = scores.sel(mode=[1])
    X_rec_1_list = eof.inverse_transform(scores_selection)
    assert isinstance(X_rec_1_list, xr.DataArray)

    # Single mode and list should be equal
    xr.testing.assert_allclose(X_rec_1, X_rec_1_list)

    # Test with all modes
    X_rec = eof.inverse_transform(scores, normalized=normalized)
    assert isinstance(X_rec, xr.DataArray)

    # Check that the reconstructed data has the same dimensions as the original data
    assert set(X_rec.dims) == set(mock_data_array.dims)

    # Reconstructed data should be close to the original data
    orig_dim_order = mock_data_array.dims
    X_rec = X_rec.transpose(*orig_dim_order)
    xr.testing.assert_allclose(mock_data_array, X_rec)


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
@pytest.mark.parametrize("engine", ["h5netcdf", "netcdf4", "zarr"])
def test_save_load(dim, mock_data_array, tmp_path, engine):
    """Test save/load methods in EOF class, ensuring that we can
    roundtrip the model and get the same results when transforming
    data."""
    skip_if_missing_engine(engine)

    original = EOF()
    original.fit(mock_data_array, dim)

    # Save the EOF model
    original.save(tmp_path / "eof", engine=engine)

    # Check that the EOF model has been saved
    assert (tmp_path / "eof").exists()

    # Recreate the model from saved file
    loaded = EOF.load(tmp_path / "eof", engine=engine)

    # Check that the params and DataContainer objects match
    assert original.get_params() == loaded.get_params()
    assert all([key in loaded.data for key in original.data])
    for key in original.data:
        if original.data._allow_compute[key]:
            assert loaded.data[key].equals(original.data[key])
        else:
            # but ensure that input data is not saved by default
            assert loaded.data[key].size <= 1
            assert loaded.data[key].attrs["placeholder"] is True

    # Test that the recreated model can be used to transform new data
    assert np.allclose(
        original.transform(mock_data_array), loaded.transform(mock_data_array)
    )

    # The loaded model should also be able to inverse_transform new data
    assert np.allclose(
        original.inverse_transform(original.scores()),
        loaded.inverse_transform(loaded.scores()),
    )


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_serialize_deserialize_dataarray(dim, mock_data_array):
    """Test roundtrip serialization when the model is fit on a DataArray."""
    model = EOF()
    model.fit(mock_data_array, dim)
    dt = model.serialize()
    rebuilt_model = EOF.deserialize(dt)
    assert np.allclose(
        model.transform(mock_data_array), rebuilt_model.transform(mock_data_array)
    )


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_serialize_deserialize_dataset(dim, mock_dataset):
    """Test roundtrip serialization when the model is fit on a Dataset."""
    model = EOF()
    model.fit(mock_dataset, dim)
    dt = model.serialize()
    rebuilt_model = EOF.deserialize(dt)
    assert np.allclose(
        model.transform(mock_dataset), rebuilt_model.transform(mock_dataset)
    )


def test_complex_input_inverse_transform(mock_complex_data_array):
    """Test that the EOF model can handle complex input data."""

    model = EOF(n_modes=128)
    model.fit(mock_complex_data_array, dim="time")
    scores = model.scores()

    # Test the inverse_transform method
    X_rec = model.inverse_transform(scores)
    assert isinstance(X_rec, xr.DataArray)
    # Check that the reconstructed data has the same dimensions as the original data
    assert set(X_rec.dims) == set(mock_complex_data_array.dims)
    # Reconstructed data should be close to the original data
    xr.testing.assert_allclose(mock_complex_data_array, X_rec)



"""
Rotated EOF analysis
========================

EOF (Empirical Orthogonal Function) analysis is commonly used in climate science, interpreting
the derived eigenvectors (EOFs) as climatic variability patterns. However, due to
the inherent orthogonality constraint in EOF analysis, the interpretation of all
but the first EOF can be problematic. Rotated EOF analysis, using optimization criteria
like Varimax and Promax, offers a solution by releasing this orthogonality constraint,
thus enabling a more accurate interpretation of variability patterns.

Both Varimax (orthogonal) and Promax (oblique) rotations result in "sparse" solutions,
meaning the EOFs become more interpretable by limiting the number of variables that
contribute to each EOF. This rotation effectively serves as a regularization method
for the EOF solution, with the strength of regularization determined by the power parameter;
the higher the value, the sparser the EOFs.

Promax rotation, with a small regularization value (i.e., power=1), reverts to Varimax
rotation. In this context, we compare the first three modes of EOF analysis: (1)
without regularization, (2) with Varimax rotation, and (3) with Promax rotation.

We'll start by loading the necessary packages and data:
"""

import matplotlib.pyplot as plt
import seaborn as sns
import xarray as xr
from cartopy.crs import PlateCarree, Robinson
from matplotlib.gridspec import GridSpec

import xeofs as xe

sns.set_context("paper")

sst = xr.tutorial.open_dataset("ersstv5")["sst"]


# %%
# Perform the actual analysis

components = []
scores = []
# (1) Standard EOF without regularization
model = xe.single.EOF(n_modes=100, standardize=True, use_coslat=True)
model.fit(sst, dim="time")
components.append(model.components())
scores.append(model.scores())
# (2) Varimax-rotated EOF analysis
rot_var = xe.single.EOFRotator(n_modes=50, power=1)
rot_var.fit(model)
components.append(rot_var.components())
scores.append(rot_var.scores())
# (3) Promax-rotated EOF analysis
rot_pro = xe.single.EOFRotator(n_modes=50, power=4)
rot_pro.fit(model)
components.append(rot_pro.components())
scores.append(rot_pro.scores())


# %%
# Create figure showing the first 6 modes for all 3 cases. While the first mode
# is very similar in all three cases the subsequent modes of the standard
# solution exhibit dipole and tripole-like patterns. Under Varimax and Promax
# rotation, these structures completely disappear suggesting that these patterns
# were mere artifacts due to the orthogonality.

proj = Robinson(central_longitude=180)
kwargs = {
    "cmap": "RdBu",
    "transform": PlateCarree(),
    "vmin": -0.03,
    "vmax": +0.03,
    "add_colorbar": False,
}

fig = plt.figure(figsize=(10, 5))
gs = GridSpec(3, 4)
ax_std = [fig.add_subplot(gs[0, i], projection=proj) for i in range(4)]
ax_var = [fig.add_subplot(gs[1, i], projection=proj) for i in range(4)]
ax_pro = [fig.add_subplot(gs[2, i], projection=proj) for i in range(4)]

for i, (a0, a1, a2) in enumerate(zip(ax_std, ax_var, ax_pro)):
    mode = i + 1
    a0.coastlines(color=".5")
    a1.coastlines(color=".5")
    a2.coastlines(color=".5")
    components[0].sel(mode=mode).plot(ax=a0, **kwargs)
    components[1].sel(mode=mode).plot(ax=a1, **kwargs)
    components[2].sel(mode=mode).plot(ax=a2, **kwargs)

title_kwargs = dict(rotation=90, va="center", weight="bold")
ax_std[0].text(-0.1, 0.5, "Standard", transform=ax_std[0].transAxes, **title_kwargs)
ax_var[0].text(-0.1, 0.5, "Varimax", transform=ax_var[0].transAxes, **title_kwargs)
ax_pro[0].text(-0.1, 0.5, "Promax", transform=ax_pro[0].transAxes, **title_kwargs)

plt.tight_layout()
plt.savefig("rotated_eof.jpg", dpi=200)



import numpy as np
from dask.array import Array as DaskArray  # type: ignore
from dask.array.linalg import svd_compressed


def _promax(
    X: np.ndarray,
    power: int = 1,
    max_iter: int = 1000,
    rtol: float = 1e-8,
    compute: bool = True,
):
    """
    Perform (oblique) Promax rotation.

    This implementation also works for complex numbers.

    Parameters
    ----------
    X : np.ndarray
        2D matrix to be rotated. Must have shape ``p x m`` containing
        p features and m modes.
    power : int
        Rotation parameter defining the power the Varimax solution is raised
        to. For ``power=1``, this is equivalent to the Varimax solution
        (the default is 1).
    max_iter: int
        Maximum number of iterations for finding the rotation matrix
        (the default is 1000).
    rtol:
        The relative tolerance for the rotation process to achieve
        (the default is 1e-8).
    compute: bool
        Whether to eagerly compute the rotation matrix. If ``True``, then
        the rotation algorithm will check for convergence at each iteration
        and stop once reached. If ``False``, then the rotation algorithm
        can build the matrix lazily, but it will not check for convergence
        and will run all of ``max_iter``.

    Returns
    -------
    Xrot : np.ndarray
        2D matrix containing the rotated modes.
    rot_mat : np.ndarray
        Rotation matrix of shape ``m x m`` with m being number of modes.
    phi : np.ndarray
        Correlation matrix of PCs of shape ``m x m`` with m being number
        of modes. For Varimax solution (``power=1``), the correlation matrix
        is diagonal i.e. the modes are uncorrelated.

    """
    X = X.copy()

    # Perform varimax rotation
    X, rot_mat = _varimax(X=X, max_iter=max_iter, rtol=rtol, compute=compute)

    # Pre-normalization by communalities (sum of squared rows)
    h = np.sqrt(np.sum(X * X.conj(), axis=1))
    # Add a stabilizer to avoid zero communalities
    eps = np.finfo(X.dtype).eps
    X = (1.0 / (h + eps))[:, np.newaxis] * X

    # Max-normalisation of columns
    Xnorm = X / np.max(abs(X), axis=0)

    # "Procustes" equation
    P = Xnorm * np.abs(Xnorm) ** (power - 1)

    # Fit linear regression model of "Procrustes" equation
    # see Richman 1986 for derivation
    L = np.linalg.inv(X.conj().T @ X) @ X.conj().T @ P

    # calculate diagonal of inverse square
    try:
        sigma_inv = np.diag(np.diag(np.linalg.inv(L.conj().T @ L)))
    except np.linalg.LinAlgError:
        sigma_inv = np.diag(np.diag(np.linalg.pinv(L.conj().T @ L)))

    # transform and calculate inner products
    L = L @ np.sqrt(sigma_inv)
    Xrot = X @ L

    # Post-normalization based on Kaiser
    Xrot = h[:, np.newaxis] * Xrot

    rot_mat = rot_mat @ L

    # Correlation matrix
    L_inv = np.linalg.inv(L)
    phi = L_inv @ L_inv.conj().T

    return Xrot, rot_mat, phi


def _varimax(
    X: np.ndarray,
    gamma: float = 1,
    max_iter: int = 1000,
    rtol: float = 1e-8,
    compute: bool = True,
):
    """
    Perform (orthogonal) Varimax rotation.

    This implementation also works for complex numbers.

    Parameters
    ----------
    X : np.ndarray
        2D matrix to be rotated containing features as rows and modes as
        columns.
    gamma : float
        Parameter which determines the type of rotation performed: varimax (1),
        quartimax (0). Other values are possible. The default is 1.
    max_iter : int
        Number of iterations performed. The default is 1000.
    rtol : float
        Relative tolerance at which iteration process terminates.
        The default is 1e-8.
    compute: bool
        Whether to eagerly compute the rotation matrix. If ``True``, then
        the rotation algorithm will check for convergence at each iteration
        and stop once reached. If ``False``, then the rotation algorithm
        can build the matrix lazily, but it will not check for convergence
        and will run all of ``max_iter``.

    Returns
    -------
    Xrot : np.ndarray
        Rotated matrix with same dimensions as X.
    R : array-like
        Rotation matrix of shape ``(n_rot x n_rot)``

    """
    X = X.copy()
    n_samples, n_modes = X.shape

    if isinstance(X, DaskArray):
        # Use svd_compressed if dask to allow chunking in both dimensions
        svd_func = svd_compressed
        svd_args = (n_modes,)
    else:
        svd_func = np.linalg.svd
        svd_args = ()

    if n_modes < 2:
        err_msg = "Cannot rotate {:} modes (columns), but must be 2 or more."
        err_msg = err_msg.format(n_modes)
        raise ValueError(err_msg)

    # Initialize rotation matrix
    R = np.eye(n_modes)

    # Normalize the matrix using square root of the sum of squares (Kaiser)
    h = np.sqrt(np.sum(X * X.conj(), axis=1))
    # A = np.diag(1./h) @ A

    # Add a stabilizer to avoid zero communalities
    eps = np.finfo(X.dtype).eps
    X = (1.0 / (h + eps))[:, np.newaxis] * X

    # Seek for rotation matrix based on varimax criteria
    delta = 0.0
    XH = X.conj().T
    alpha = gamma / n_samples
    for i in range(max_iter):
        delta_old = delta
        basis = X @ R

        basis2 = basis * basis.conj()
        W = np.sum(basis2, axis=0)
        transformed = XH @ (basis * (basis2 - (alpha * W)))
        U, svals, VT = svd_func(transformed, *svd_args)
        R = U @ VT
        delta = np.sum(svals)
        if compute and (abs(delta - delta_old) / delta) < rtol:
            break

    if compute and (abs(delta - delta_old) / delta) > rtol:
        raise RuntimeError("Rotation process did not converge.")

    # De-normalize
    X = h[:, np.newaxis] * X

    # Rotate
    Xrot = X @ R
    return Xrot, R



"""
Sparse PCA
========================

This example demonstrates the application of sparse PCA [1]_ to sea surface temperature data. Sparse PCA is an alternative to rotated PCA, where the components are sparse, often providing a more interpretable solution.

We replicate the analysis from the original paper [1]_, which identifies the ENSO (El Niño-Southern Oscillation) as the fourth mode, representing about 1% of the total variance. The original study focused on weekly sea surface temperatures from satellite data, whereas we use monthly data from ERSSTv5. Consequently, our results may not match exactly, but they should be quite similar.

References
----------
.. [1] Erichson, N. B. et al. Sparse Principal Component Analysis via Variable Projection. SIAM J. Appl. Math. 80, 977-1002 (2020).

"""

# Load packages and data:
import matplotlib.pyplot as plt
import xarray as xr
from cartopy.crs import EqualEarth, PlateCarree
from matplotlib.gridspec import GridSpec

import xeofs as xe

# %%
# We use sea surface temperature data from 1990 to 2017, consistent with the original paper.

sst = xr.tutorial.open_dataset("ersstv5")["sst"]
sst = sst.sel(time=slice("1990", "2017"))

# %%
# We perform sparse PCA using the `alpha` and `beta` parameters, which define the sparsity imposed by the elastic net (refer to Table 1 in the paper). In our analysis, we set `alpha` to 1e-5, as specified by the authors. Although the authors do not specify a value for `beta`, it appears that the results are not highly sensitive to this parameter. Therefore, we use the default `beta` value of 1e-4.

model = xe.single.SparsePCA(n_modes=4, alpha=1e-5)
model.fit(sst, dim="time")
expvar = model.explained_variance()
expvar_ratio = model.explained_variance_ratio()
components = model.components()
scores = model.scores()

# %%
# The explained variance fraction confirms that the fourth mode explains about 1% of the total variance, which is consistent with the original paper.

print("Explained variance: ", expvar.round(0).values)
print("Relative: ", (expvar_ratio * 100).round(1).values)

# %%
# Examining the first four modes, we clearly identify ENSO as the fourth mode.

proj = EqualEarth(central_longitude=180)
kwargs = {"cmap": "RdBu", "vmin": -0.05, "vmax": 0.05, "transform": PlateCarree()}

fig = plt.figure(figsize=(10, 12))
gs = GridSpec(4, 2, width_ratios=[1, 2])
ax0 = [fig.add_subplot(gs[i, 0]) for i in range(4)]
ax1 = [fig.add_subplot(gs[i, 1], projection=proj) for i in range(4)]

for i, (a0, a1) in enumerate(zip(ax0, ax1)):
    scores.sel(mode=i + 1).plot(ax=a0)
    a1.coastlines(color=".5")
    components.sel(mode=i + 1).plot(ax=a1, **kwargs)

    a0.set_xlabel("")

plt.tight_layout()
plt.savefig("sparse_pca.jpg")

# %%



"""
Hilbert EOF analysis
============================================

We demonstrate how to execute a Hilbert EOF analysis [1]_ [2]_
[3]_. This method extends traditional EOF analysis into the complex domain,
allowing the EOF components to have real and imaginary parts. This capability
can reveal oscillatory patterns in datasets, which are common in Earth
observations. For example, beyond typical examples like seasonal cycles, you can
think of internal waves in the ocean, or the Quasi-Biennial Oscillation in the
atmosphere.

Using monthly sea surface temperature data from 1970 to 2021 as an example, we
highlight the method's key features and address edge effects as a common
challenge.

.. [1] Rasmusson, E. M., Arkin, P. A., Chen, W.-Y. & Jalickee, J. B. Biennial
    variations in surface temperature over the United States as revealed by
    singular decomposition. Monthly Weather Review 109, 587–598 (1981).
.. [2] Barnett, T. P. Interaction of the Monsoon and Pacific Trade Wind System
    at Interannual Time Scales Part I: The Equatorial Zone. Monthly Weather
    Review 111, 756–773 (1983).
.. [3] Horel, J. Complex Principal Component Analysis: Theory and Examples. J.
    Climate Appl. Meteor. 23, 1660–1673 (1984).

Let's start by importing the necessary packages and loading the data:
"""

# %%
import matplotlib.pyplot as plt
import xarray as xr

import xeofs as xe

xr.set_options(display_expand_attrs=False)

sst = xr.tutorial.open_dataset("ersstv5").sst
sst

# %%
# We fit the ``HilbertEOF`` model directly to the raw data, retaining the seasonal
# cycle for study. The model initialization specifies the desired number of
# modes. The ``use_coslat`` parameter is set to ``True`` to adjust for grid
# convergence at the poles. While the ``HilbertEOF`` class offers padding
# options to mitigate potential edge effects, we'll begin with no padding.

kwargs = dict(n_modes=4, use_coslat=True, random_state=7)
model = xe.single.HilbertEOF(padding="none", **kwargs)

# %%
# Now, we fit the model to the data and extract the explained variance.

model.fit(sst, dim="time")
expvar = model.explained_variance()
expvar_ratio = model.explained_variance_ratio()

# %%
# Let's have a look at the explained variance of the first five modes:

expvar.round(0)


# %%
# Clearly, the first mode completely dominates and already explains a
# substantial amount of variance. If we look at the fraction of explained
# variance, we see that the first mode explains about 88.8 %.

(expvar_ratio * 100).round(1)

# %%
# In comparison to standard EOF analysis (check the corresponding example,
# S-mode), the first complex mode seems to integrate the first two standard
# modes in terms of explained variance. This makes sense as the two modes in
# standard EOF are both showing parts of an annual cycle (which are in
# quadrature) and thus the complex mode combines both of them. Let's confirm our
# hypothesis by looking at the real part the complex-valued scores:

scores = model.scores()
scores.real.plot.line(x="time", col="mode", lw=1, ylim=(-0.1, 0.1))
plt.show()


# %%
# And indeed the annual cycle is completed incorporated into the first mode,
# while the second mode shows a semi-annual cycle (mode 3 in standard EOF).
# However, mode three and four look unusual. While showing some similarity to
# ENSO (e.g. in mode 3 peaks in 1982, 1998 and 2016), they exhibit a "running
# away" behaviour towards the boundaries of the time series. This a common issue
# in Hilbert EOF analysis which is based on the Hilbert transform (a
# convolution) that suffers from the absence of information at the time series
# boundaries. One way to mitigate this is to artificially extend the time series
# also known as *padding*. In ``xeofs``, you can enable such a padding by
# setting the ``padding`` parameter to ``"exp"`` which will extent the
# boundaries by an exponential decaying function. The ``decay_factor`` parameter
# controls the decay rate of the exponential function measured in multiples of
# the time series length. Let's see how the decay parameter impacts the results:

model_ext = xe.single.HilbertEOF(padding="exp", decay_factor=0.01, **kwargs)
model_ext.fit(sst, dim="time")
scores_ext = model_ext.scores().sel(mode=slice(1, 4))

scores_ext.real.plot.line(x="time", col="mode", lw=1, ylim=(-0.1, 0.1))
plt.show()

# %%
# And indeed, padding the time series effectively reduced the artifacts at the
# boundaries. Lastly, we examine the complex component amplitudes and phases.

comp_amps = model.components_amplitude()
comp_amps.plot(col="mode", vmin=0, vmax=0.025)
plt.show()

# %%
# The component phases of the first mode clearly show the seasonal cycle as the
# northern and southern hemisphere are phase shifted by 180 degrees (white and
# black). Note the blueish regions in the central East Pacific and Indian Ocean
# which indicate a phase shift of 90 degrees compared to the main annual cycle.
# This is in agreement with mode 3 of the standard EOF analysis.

comp_phases = model.components_phase()
comp_phases.plot(col="mode", cmap="twilight")
plt.show()

# %%



# %%
import xarray as xr
import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import matplotx
from matplotlib.colors import LogNorm

# Load timings
# =============================================================================
timings = xr.open_dataset("timings.nc").sortby("n_features").sortby("n_samples")
xeofs_timings = (
    timings[["xeofs", "xeofs_dask"]].to_array("package").min(("run", "package"))
)
eofs_timings = (
    timings[["eofs", "eofs_dask"]].to_array("package").min(("run", "package"))
)
ratio = xeofs_timings / eofs_timings
# %%


class MidPointLogNorm(LogNorm):
    def __init__(self, vmin=None, vmax=None, midpoint=None, clip=False):
        LogNorm.__init__(self, vmin=vmin, vmax=vmax, clip=clip)
        self.midpoint = midpoint

    def __call__(self, value, clip=None):
        result, is_scalar = self.process_value(value)
        x, y = (
            [np.log(self.vmin), np.log(self.midpoint), np.log(self.vmax)],
            [
                0,
                0.5,
                1,
            ],
        )
        return np.ma.array(np.interp(np.log(value), x, y), mask=result.mask, copy=False)


def to_scientific_notation(arr):
    return np.array([f"$10^{len(str(x))-1}$" if x != 0 else "0" for x in arr])


# %%
sns.set_context("paper")
sns.set_style("ticks")


def create_figure(path=None):
    fig, axes = plt.subplots(1, 2, figsize=(10, 4))

    ticks = [0.1, 1, 10, 60]
    ticklabels = ["100 ms", "1 s", "10 s", "1 min"]
    sns.heatmap(
        timings["xeofs_dask"].min("run"),
        ax=axes[0],
        cmap="viridis",
        norm=LogNorm(vmin=0.1, vmax=60),
        cbar_kws=dict(ticks=ticks),
    )
    axes[0].collections[0].colorbar.ax.yaxis.set_ticks([], minor=True)
    axes[0].collections[0].colorbar.set_ticklabels([t for t in ticklabels])
    xcoords = timings["xeofs"]["n_features"].values
    ycoords = timings["xeofs"]["n_samples"].values
    xticklabels = to_scientific_notation(xcoords)
    yticklabels = to_scientific_notation(ycoords)
    xticks = np.arange(0.5, len(xcoords) + 0.5, 3)
    yticks = np.arange(0.5, len(ycoords) + 0.5, 3)
    axes[0].xaxis.set_ticks(xticks)
    axes[0].yaxis.set_ticks(yticks)
    axes[0].xaxis.set_ticklabels(xticklabels[::3])
    axes[0].yaxis.set_ticklabels(yticklabels[::3], rotation=0)
    axes[0].set_xlabel("Number of features")
    axes[0].set_ylabel("Number of samples")
    axes[0].set_title("A | Runtime xeofs", loc="left")

    ticks = [0.001, 0.01, 0.1, 1, 10, 100, 1000]
    ticklabels = [
        "x1000",
        "x100",
        "x10",
        "same",
        "x10",
        "x100",
        "x1000",
    ]

    sns.heatmap(
        ratio,
        ax=axes[1],
        cmap="coolwarm",
        norm=MidPointLogNorm(vmin=0.001, vmax=1000, midpoint=1),
        cbar_kws=dict(ticks=ticks),
    )
    axes[1].collections[0].colorbar.ax.yaxis.set_ticks([], minor=True)
    axes[1].collections[0].colorbar.set_ticklabels([t for t in ticklabels])
    axes[1].collections[0].colorbar.set_label(
        "Speed\nGain", rotation=0, labelpad=40, va="center", ha="right"
    )
    xcoords = timings["xeofs"]["n_features"].values
    ycoords = timings["xeofs"]["n_samples"].values
    xticklabels = to_scientific_notation(xcoords)
    yticklabels = to_scientific_notation(ycoords)
    xticks = np.arange(0.5, len(xcoords) + 0.5, 3)
    yticks = np.arange(0.5, len(ycoords) + 0.5, 3)
    axes[1].xaxis.set_ticks(xticks)
    axes[1].yaxis.set_ticks(yticks)
    axes[1].xaxis.set_ticklabels(xticklabels[::3])
    axes[1].yaxis.set_ticklabels(yticklabels[::3], rotation=0)
    axes[1].set_xlabel("Number of features")
    axes[1].set_ylabel("Number of samples")
    axes[1].text(1.35, 1, "eofs", transform=axes[1].transAxes, ha="left", va="center")
    axes[1].text(1.35, 0, "xeofs", transform=axes[1].transAxes, ha="left", va="center")
    x = timings.n_features.values
    y = timings.n_samples.values
    xx, yy = np.meshgrid(np.arange(0.5, len(x) + 0.5), np.arange(0.5, len(y) + 0.5))
    axes[1].contour(
        xx,
        yy,
        timings["nbytes"],
        levels=[3e6],
        linestyles="--",
        colors="k",
        linewidths=0.5,
    )
    axes[1].set_title("B | Speed gain eofs vs xeofs", loc="left")
    plt.tight_layout()
    if path is not None:
        fig.savefig(path, dpi=200)
    plt.show()


# %%
create_figure("timings_light.png")
with plt.style.context(matplotx.styles.aura["dark"]):
    create_figure("timings_dark.png")
# %%



from .cross import MCA, HilbertMCA, HilbertMCARotator, MCARotator
from .single import EOF, EOFRotator, HilbertEOF, HilbertEOFRotator


class RotatorFactory:
    """Factory class for creating rotators.

    Parameters
    ----------
    n_rot : int
        Number of modes to be rotated.
    power : int
        Defines the power of Promax rotation. Choosing ``power=1`` equals
        a Varimax solution (the default is 1).
    max_iter : int
        Number of maximal iterations for obtaining the rotation matrix
        (the default is 1000).
    rtol : float
        Relative tolerance to be achieved for early stopping the iteration
        process (the default is 1e-8).

    """

    def __init__(self, **kwargs):
        self.params = kwargs
        self._valid_types = (EOF, HilbertEOF, MCA, HilbertMCA)

    def create_rotator(
        self, model: EOF | HilbertEOF | MCA | HilbertMCA
    ) -> EOFRotator | HilbertEOFRotator | MCARotator | HilbertMCARotator:
        """Create a rotator for the given model.

        Parameters
        ----------
        model : xeofs model
            Model to be rotated.

        Returns
        -------
        xeofs Rotator
            Rotator for the given model.
        """
        # We need to check the type of the model instead of isinstance because
        # of inheritance.
        if type(model) is EOF:
            return EOFRotator(**self.params)
        elif type(model) is HilbertEOF:
            return HilbertEOFRotator(**self.params)
        elif type(model) is MCA:
            return MCARotator(**self.params)
        elif type(model) is HilbertMCA:
            return HilbertMCARotator(**self.params)
        else:
            err_msg = f"Invalid model type. Valid types are {self._valid_types}."
            raise TypeError(err_msg)



import numpy as np
import xarray as xr
from typing_extensions import Self

from ..linalg.svd import SVD
from ..utils.data_types import (
    DataArray,
    Dims,
    DimsList,
)
from ..utils.sanity_checks import assert_single_dataarray
from .transformer import Transformer


class PCA(Transformer):
    """Transform data into reduced PC space.

    For large number of features, use PCA to reduce the dimensionality of the data before whitening.

    Note that for ``alpha=1`` (no whiteneing) and ``use_pca=False``, it just becomes the identity transformation.

    Parameters
    ----------
    n_modes: int | float | str, default="all"
        If int, number of components to keep. If float, fraction of variance to keep. If `n_modes="all"`, keep all components.
    use_pca: bool, default=True
        Whether or not to use PCA to reduce the dimensionality of the data. If False, perform identity transformation.
    init_rank_reduction: float, default=0.3
        Used only when `n_modes` is given as a float. Specifiy the initial PCA rank reduction before truncating the solution to the desired fraction of explained variance. Must be in the half open interval ]0, 1]. Lower values will speed up the computation.
    compute_eagerly: bool, default=False
        Whether to perform eager or lazy computation.
    sample_name: str, default="sample"
        Name of the sample dimension.
    feature_name: str, default="feature"
        Name of the feature dimension.
    random_state: np.random.Generator | int | None, default=None
        Random seed for reproducibility.
    solver_kwargs: dict
        Additional keyword arguments for the SVD solver.

    """

    def __init__(
        self,
        n_modes: int | float | str = "all",
        use_pca: bool = True,
        init_rank_reduction: float = 0.3,
        compute_eagerly: bool = False,
        sample_name: str = "sample",
        feature_name: str = "feature",
        random_state: np.random.Generator | int | None = None,
        solver_kwargs: dict = {},
    ):
        super().__init__(sample_name, feature_name)

        self.use_pca = use_pca
        self.n_modes = n_modes
        self.init_rank_reduction = init_rank_reduction
        self.compute_eagerly = compute_eagerly
        self.random_state = random_state
        self.solver_kwargs = solver_kwargs

        # Check whether Whitener is identity transformation
        self.is_identity = not use_pca

    def _sanity_check_input(self, X) -> None:
        assert_single_dataarray(X)

        if len(X.dims) != 2:
            raise ValueError("Input DataArray must have shape 2")

        if X.dims != (self.sample_name, self.feature_name):
            raise ValueError(
                "Input DataArray must have dimensions ({:}, {:})".format(
                    self.sample_name, self.feature_name
                )
            )

    def _get_n_modes(self, X: DataArray) -> int | float:
        if isinstance(self.n_modes, str):
            if self.n_modes == "all":
                return min(X.shape)
            else:
                raise ValueError("`n_modes` must be an integer, float or 'all'")
        else:
            return self.n_modes

    def get_serialization_attrs(self) -> dict:
        return dict(
            V=self.V,
            use_pca=self.use_pca,
        )

    def fit(
        self,
        X: DataArray,
        sample_dims: Dims | None = None,
        feature_dims: DimsList | None = None,
    ) -> Self:
        self._sanity_check_input(X)

        if self.use_pca:
            # In case of "all" modes to the rank of the input data
            self.n_modes = self._get_n_modes(X)

            svd = SVD(
                n_modes=self.n_modes,
                init_rank_reduction=self.init_rank_reduction,
                compute=self.compute_eagerly,
                random_state=self.random_state,
                sample_name=self.sample_name,
                feature_name=self.feature_name,
                **self.solver_kwargs,
            )
            _, _, self.V = svd.fit_transform(X)

        else:
            self.V = xr.DataArray(1, name="identity")

        return self

    def transform(self, X: DataArray) -> DataArray:
        """Transform new data into the PC space."""

        self._sanity_check_input(X)
        if self.use_pca:
            transformed = xr.dot(X, self.V, dims=self.feature_name)
            transformed.name = X.name
            return transformed.rename({"mode": self.feature_name})
        else:
            return X

    def fit_transform(
        self,
        X: DataArray,
        sample_dims: Dims | None = None,
        feature_dims: DimsList | None = None,
    ) -> DataArray:
        return self.fit(X, sample_dims, feature_dims).transform(X)

    def inverse_transform_data(self, X: DataArray) -> DataArray:
        """Transform 2D data (sample x feature) from PC space back into original space."""
        if self.use_pca:
            X = X.rename({self.feature_name: "mode"})
            return xr.dot(X, self.V.conj().T, dims="mode")
        else:
            return X

    def transform_components(self, X: DataArray) -> DataArray:
        """Transform 2D components (feature x mode) into PC space."""

        if self.use_pca:
            dummy_dim = "dummy_dim"
            Tinv = self.V.conj().T
            Tinv = Tinv.rename({"mode": dummy_dim})
            transformed = xr.dot(Tinv, X, dims=self.feature_name)
            return transformed.rename({dummy_dim: self.feature_name})
        else:
            return X

    def inverse_transform_components(self, X: DataArray) -> DataArray:
        """Transform 2D components (feature x mode) from PC space back into original space."""

        if self.use_pca:
            dummy_dim = "dummy_dim"
            comps_pc_space = X.rename({self.feature_name: dummy_dim})
            V = self.V
            V = V.rename({"mode": dummy_dim})
            return xr.dot(V, comps_pc_space, dims=dummy_dim)
        else:
            return X

    def inverse_transform_scores(self, X: DataArray) -> DataArray:
        """Transform 2D scores (sample x mode) from whitened PC space back into original space."""

        return X

    def inverse_transform_scores_unseen(self, X: DataArray) -> DataArray:
        """Transform unseen 2D scores (sample x mode) from whitened PC space back into original space."""

        return X



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

from xeofs import __version__

sys.path.insert(0, os.path.abspath(".."))


# # prevent circular imports...
# import sphinx.builders.html
# import sphinx.builders.latex
# import sphinx.builders.texinfo
# import sphinx.builders.text
# import sphinx.ext.autodoc

# -- Project information -----------------------------------------------------

project = "xeofs"
copyright = "2022 - 2024, Niclas Rieger"
author = "Niclas Rieger"

# The short X.Y version
# version = importlib.metadata.version('xeofs')
# The full version, including alpha/beta/rc tags
release = __version__


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.napoleon",
    # "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx_gallery.gen_gallery",
    "sphinx_design",
    "nbsphinx",
    "sphinx_copybutton",
    "myst_parser",  # markdown support
]
autosummary_generate = True  # Turn on sphinx.ext.autosummary


add_module_names = False
toc_object_entries_show_parents = "hide"
show_title_parents = False

# Sphinx-gallery stuff
sphinx_gallery_conf = {
    "examples_dirs": "content/user_guide/examples",  # path to your example scripts
    "gallery_dirs": "content/user_guide/auto_examples",  # path to where to save gallery generated output
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    # exclude .py and .ipynb files in auto_examples generated by sphinx-gallery
    # this is to prevent sphinx from complaining about duplicate source files
    "content/user_guide/auto_examples/**/*.py",
    "content/user_guide/auto_examples/**/*.ipynb",
]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"

html_context = {
    # ...
    "default_mode": "dark"
}

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "show_nav_level": 1,
    "icon_links": [
        {
            # Label for this link
            "name": "GitHub",
            # URL where the link will redirect
            "url": "https://github.com/xarray-contrib/xeofs",  # required
            # Icon class (if "type": "fontawesome"), or path to local image (if "type": "local")
            "icon": "fa-brands fa-square-github",
            # The type of image to be used (see below for details)
            "type": "fontawesome",
        }
    ],
}

html_logo = "./logos/xeofs_logo_icon.png"
html_favicon = "./logos/favicon.ico"
# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_css_files = ["custom.css"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "xeofsdoc"


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, "xeofs.tex", "xeofs Documentation", "Niclas Rieger", "manual"),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "xeofs", "xeofs Documentation", [author], 1)]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "xeofs",
        "xeofs Documentation",
        author,
        "xeofs",
        "Comprehensive EOF analysis in Python.",
        "Miscellaneous",
    ),
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ["search.html"]


# Remove parents from titles in all .rst files
def shorten_titles(app):
    # Recursively crawl through source directory and shorten titles in .rst files
    def crawl_source_shorten_titles(path):
        # List files in directory
        for file_name in os.listdir(path):
            # Build path to file
            file_path = os.path.join(path, file_name)

            # Recursively crawl to next directory level
            if os.path.isdir(file_path):
                crawl_source_shorten_titles(file_path)

            # Modify .rst source file title
            else:
                _, extension = os.path.splitext(file_path)
                if extension == ".rst":
                    # Read file, modify title, write back to file
                    with open(file_path, "r") as file:
                        lines = file.readlines()
                    lines[0] = lines[0].split(".")[-1]
                    lines[1] = ("=" * (len(lines[0]) - 1)) + "\n"
                    print(f"Shortened title in {file_path}")
                    with open(file_path, "w") as file:
                        file.writelines(lines)

    path = os.path.join(
        os.path.abspath(".."), "docs/content/api_reference/_autosummary"
    )
    crawl_source_shorten_titles(path)


# Connect to sphinx events (see https://www.sphinx-doc.org/en/master/extdev/event_callbacks.html#events)
# to shorten titles in all .rst files after the sphinx autosummary extension has run
# (otherwise the titles would be overwritten by the autosummary extension)
def setup(app):
    app.connect("builder-inited", shorten_titles)



import numpy as np
import pytest
import xarray as xr

from xeofs.single import POP

from ...utilities import skip_if_missing_engine


def test_init():
    """Tests the initialization of the POP class"""
    pop = POP(n_modes=5, standardize=True, use_coslat=True)

    # Assert preprocessor has been initialized
    assert hasattr(pop, "_params")
    assert hasattr(pop, "preprocessor")
    assert hasattr(pop, "pca")


def test_fit(mock_data_array):
    pop = POP()
    pop.fit(mock_data_array, "time")


def test_eigenvalues(mock_data_array):
    pop = POP()
    pop.fit(mock_data_array, "time")

    eigenvalues = pop.eigenvalues()
    assert isinstance(eigenvalues, xr.DataArray)


def test_damping_times(mock_data_array):
    pop = POP()
    pop.fit(mock_data_array, "time")

    times = pop.damping_times()
    assert isinstance(times, xr.DataArray)


def test_periods(mock_data_array):
    pop = POP()
    pop.fit(mock_data_array, "time")

    periods = pop.periods()
    assert isinstance(periods, xr.DataArray)


def test_components(mock_data_array):
    """Tests the components method of the POP class"""
    sample_dim = ("time",)
    pop = POP()
    pop.fit(mock_data_array, sample_dim)

    # Test components method
    components = pop.components()
    feature_dims = tuple(set(mock_data_array.dims) - set(sample_dim))
    assert isinstance(components, xr.DataArray), "Components is not a DataArray"
    assert set(components.dims) == set(
        ("mode",) + feature_dims
    ), "Components does not have the right feature dimensions"


def test_scores(mock_data_array):
    """Tests the scores method of the POP class"""
    sample_dim = ("time",)
    pop = POP()
    pop.fit(mock_data_array, sample_dim)

    # Test scores method
    scores = pop.scores()
    assert isinstance(scores, xr.DataArray), "Scores is not a DataArray"
    assert set(scores.dims) == set(
        (sample_dim + ("mode",))
    ), "Scores does not have the right dimensions"


def test_transform(mock_data_array):
    """Test projecting new unseen data onto the POPs"""
    dim = ("time",)
    data = mock_data_array.isel({dim[0]: slice(1, None)})
    new_data = mock_data_array.isel({dim[0]: slice(0, 1)})

    # Create a xarray DataArray with random data
    model = POP(n_modes=2, solver="full")
    model.fit(data, dim)
    scores = model.scores()

    # Project data onto the components
    projections = model.transform(data)

    # Check that the projection has the right dimensions
    assert projections.dims == scores.dims, "Projection has wrong dimensions"  # type: ignore

    # Check that the projection has the right data type
    assert isinstance(projections, xr.DataArray), "Projection is not a DataArray"

    # Check that the projection has the right name
    assert projections.name == "scores", "Projection has wrong name: {}".format(
        projections.name
    )

    # Check that the projection's data is the same as the scores
    np.testing.assert_allclose(
        scores.sel(mode=slice(1, 3)), projections.sel(mode=slice(1, 3)), rtol=1e-3
    )

    # Project unseen data onto the components
    new_projections = model.transform(new_data)

    # Check that the projection has the right dimensions
    assert new_projections.dims == scores.dims, "Projection has wrong dimensions"  # type: ignore

    # Check that the projection has the right data type
    assert isinstance(new_projections, xr.DataArray), "Projection is not a DataArray"

    # Check that the projection has the right name
    assert new_projections.name == "scores", "Projection has wrong name: {}".format(
        new_projections.name
    )

    # Ensure that the new projections are not NaNs
    assert np.all(new_projections.notnull().values), "New projections contain NaNs"


def test_inverse_transform(mock_data_array):
    """Test inverse_transform method in POP class."""

    dim = ("time",)
    # instantiate the POP class with necessary parameters
    pop = POP(n_modes=20, standardize=True)

    # fit the POP model
    pop.fit(mock_data_array, dim=dim)
    scores = pop.scores()

    # Test with single mode
    scores_selection = scores.sel(mode=1)
    X_rec_1 = pop.inverse_transform(scores_selection)
    assert isinstance(X_rec_1, xr.DataArray)

    # Test with single mode as list
    scores_selection = scores.sel(mode=[1])
    X_rec_1_list = pop.inverse_transform(scores_selection)
    assert isinstance(X_rec_1_list, xr.DataArray)

    # Single mode and list should be equal
    xr.testing.assert_allclose(X_rec_1, X_rec_1_list)

    # Test with all modes
    X_rec = pop.inverse_transform(scores)
    assert isinstance(X_rec, xr.DataArray)

    # Check that the reconstructed data has the same dimensions as the original data
    assert set(X_rec.dims) == set(mock_data_array.dims)


@pytest.mark.parametrize("engine", ["h5netcdf", "zarr"])
def test_save_load(mock_data_array, tmp_path, engine):
    """Test save/load methods in POP class, ensuring that we can
    roundtrip the model and get the same results when transforming
    data."""
    # NOTE: netcdf4 does not support complex data types, so we use only zarr here
    skip_if_missing_engine(engine)

    dim = "time"
    original = POP()
    original.fit(mock_data_array, dim)

    # Save the POP model
    original.save(tmp_path / "pop", engine=engine)

    # Check that the POP model has been saved
    assert (tmp_path / "pop").exists()

    # Recreate the model from saved file
    loaded = POP.load(tmp_path / "pop", engine=engine)

    # Check that the params and DataContainer objects match
    assert original.get_params() == loaded.get_params()
    assert all([key in loaded.data for key in original.data])
    for key in original.data:
        if original.data._allow_compute[key]:
            assert loaded.data[key].equals(original.data[key])
        else:
            # but ensure that input data is not saved by default
            assert loaded.data[key].size <= 1
            assert loaded.data[key].attrs["placeholder"] is True

    # Test that the recreated model can be used to transform new data
    assert np.allclose(
        original.transform(mock_data_array), loaded.transform(mock_data_array)
    )

    # The loaded model should also be able to inverse_transform new data
    assert np.allclose(
        original.inverse_transform(original.scores()),
        loaded.inverse_transform(loaded.scores()),
    )


def test_serialize_deserialize_dataarray(mock_data_array):
    """Test roundtrip serialization when the model is fit on a DataArray."""
    dim = "time"
    model = POP()
    model.fit(mock_data_array, dim)
    dt = model.serialize()
    rebuilt_model = POP.deserialize(dt)
    assert np.allclose(
        model.transform(mock_data_array), rebuilt_model.transform(mock_data_array)
    )


def test_serialize_deserialize_dataset(mock_dataset):
    """Test roundtrip serialization when the model is fit on a Dataset."""
    dim = "time"
    model = POP()
    model.fit(mock_dataset, dim)
    dt = model.serialize()
    rebuilt_model = POP.deserialize(dt)
    assert np.allclose(
        model.transform(mock_dataset), rebuilt_model.transform(mock_dataset)
    )



"""
Sparse PCA via Variable Projection [1]_.

This module provides an implementation of Sparse PCA via Variable Projection,
based heavily on the code from the Ristretto library (https://github.com/erichson/ristretto).

We have made modifications to adapt the algorithm to be used with delayed Dask objects.

This code is licensed under the MIT License. The original library is licensed under the GPL-3.0 License.

References
----------

.. [1] Erichson, N. B. et al. Sparse Principal Component Analysis via Variable Projection. SIAM J. Appl. Math. 80, 977-1002 (2020).

"""

import dask.array.core
import dask.array.linalg
import dask.array.random
import numpy as np
import scipy
from sklearn.utils import check_random_state

from ...utils.data_types import DaskArray


def random_gaussian_map(A, k, random_state):
    """generate random gaussian map

    Parameters
    ----------
    A : array_like, shape `(n, p)`.
        Input array.
    k : integer
        Target rank.
    random_state : RandomState instance
        Random number generator.

    Returns
    -------
    Omega : array_like, `(p, k)`.
        Random gaussian matrix.

    """

    # TODO: adapt for complex-valued data

    if isinstance(A, DaskArray):
        Omega = dask.array.random.standard_normal(
            size=(A.shape[1], k), chunks=(A.chunks[1], -1)
        )
        return Omega.astype(A.dtype)
    else:
        return random_state.standard_normal(size=(A.shape[1], k)).astype(A.dtype)


def johnson_lindenstrauss(A, k, random_state=None):
    """Johnson-Lindenstrauss Random Projection.

    Parameters
    ----------
    A : array_like, shape `(n, p)`.
        Input array.
    k : integer
        Target rank.
    random_state : integer, RandomState instance or None
        If integer, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used by np.random.

    Returns
    -------
    array_like, `(n, k)`.
        Projected matrix.

    """
    # TODO: adapt for complex-valued data
    random_state = check_random_state(random_state)

    if A.ndim != 2:
        raise ValueError("A must be a 2D array, not %dD" % A.ndim)

    # construct gaussian random matrix
    Omega = random_gaussian_map(A, k, random_state)

    # project A onto Omega
    return A.dot(Omega)


def orthonormalize(A, overwrite_a=True, check_finite=False):
    """orthonormalize the columns of A via QR decomposition

    Parameters
    ----------
    A : array_like, shape `(n, p)`.
        Input array.
    overwrite_a : bool, optional
        Whether to overwrite data in A (may improve performance).
    check_finite : bool, optional
        Whether to check that the input matrix contains only finite numbers.

    Returns
    -------
    Q : array_like, `(n, p)`.
        Orthonormal basis matrix.

    """
    # TODO: adapt for complex-valued data
    if isinstance(A, DaskArray):
        # NOTE: We expect A  to be tall-and-skinny matrix that is chunked along rows
        # This means our input data is assumed to be chunked along the feature dimensions
        try:
            Q, _ = dask.array.linalg.tsqr(A)
        except ValueError:
            raise ValueError(
                "Data not chunked correctly. Ensure that the data is chunked along the feature dimensions."
            )
    else:
        # NOTE: for A(n, p) 'economic' returns Q(n, k), R(k, p) where k is min(n, p)
        # TODO: when does overwrite_a even work? (fortran?)
        QR = scipy.linalg.qr(
            A,
            overwrite_a=overwrite_a,
            check_finite=check_finite,
            mode="economic",
            pivoting=False,
        )
        Q = QR[0]
    return Q


def perform_subspace_iterations(A, Q, n_iter=2):
    """perform subspace iterations on Q

    Parameters
    ----------
    A : array_like, shape `(n, p)`.
        Input array.
    Q : array_like, shape `(n, k)`.
        Orthonormal basis matrix.
    n_iter : integer, default: 2.
        Number of subspace iterations.

    Returns
    -------
    Q : array_like, `(n, k)`.
        Orthonormal basis matrix.

    """
    # TODO: adapt for complex-valued data
    # orthonormalize Y, overwriting
    Q = orthonormalize(Q)

    # perform subspace iterations
    for _ in range(n_iter):
        Z = orthonormalize(A.T.dot(Q))
        Q = orthonormalize(A.dot(Z))

    return Q


def conjugate_transpose(A):
    """Performs conjugate transpose of A"""
    if np.iscomplexobj(A):
        return A.conj().T
    return A.T


def _compute_rqb(A, rank, oversample=10, n_subspace=2, random_state=None):
    """Randomized QB Decomposition.

    Randomized algorithm for computing the approximate low-rank QB
    decomposition of a rectangular `(n, p)` matrix `A`, with target rank
    `rank << min{n, p}`. The input matrix is factored as `A = Q * B`.

    The quality of the approximation can be controlled via the oversampling
    parameter `oversample` and `n_subspace` which specifies the number of
    subspace iterations.

    Parameters
    ----------
    A : array_like, shape `(n, p)`.
        Input array.
    rank : integer
        Target rank. Best if `rank << min{n, p}`
    oversample : integer, optional (default: 10)
        Controls the oversampling of column space. Increasing this parameter
        may improve numerical accuracy.
    n_subspace : integer, default: 2.
        Parameter to control number of subspace iterations. Increasing this
        parameter may improve numerical accuracy. Every additional subspace
        iterations requires an additional full pass over the data matrix.
    random_state : integer, RandomState instance or None
        If integer, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used by np.random.

    Returns
    -------
    Q:  array_like, shape `(n, rank + oversample)`.
        Orthonormal basis matrix.
    B : array_like, shape `(rank + oversample, p)`.
        Smaller matrix.

    """
    # TODO: adapt for complex-valued data
    Q = johnson_lindenstrauss(A, rank + oversample, random_state=random_state)

    if n_subspace > 0:
        Q = perform_subspace_iterations(A, Q, n_iter=n_subspace)
    else:
        Q = orthonormalize(Q)

    # Project the data matrix a into a lower dimensional subspace
    B = conjugate_transpose(Q).dot(A)

    return Q, B


def compute_rqb(A, rank, oversample=20, n_subspace=2, n_blocks=1, random_state=None):
    """Randomized QB Decomposition.

    Randomized algorithm for computing the approximate low-rank QB
    decomposition of a rectangular `(n, p)` matrix `A`, with target rank
    `rank << min{n, p}`. The input matrix is factored as `A = Q * B`.

    The quality of the approximation can be controlled via the oversampling
    parameter `oversample` and `n_subspace` which specifies the number of
    subspace iterations.

    Parameters
    ----------
    A : array_like, shape `(n, p)`.
        Input array.

    rank : integer
        Target rank. Best if `rank << min{n, p}`

    oversample : integer, optional (default: 10)
        Controls the oversampling of column space. Increasing this parameter
        may improve numerical accuracy.

    n_subspace : integer, default: 2.
        Parameter to control number of subspace iterations. Increasing this
        parameter may improve numerical accuracy. Every additional subspace
        iterations requires an additional full pass over the data matrix.

    n_blocks : integer, default: 1.
        If `n_blocks > 1` a column blocked QB decomposition procedure will be
        performed. A larger number requires less fast memory, while it
        leads to a higher computational time.

    random_state : integer, RandomState instance or None, optional (default `None`)
        If integer, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used by np.random.

    Returns
    -------
    Q:  array_like, shape `(n, rank + oversample)`.
        Orthonormal basis matrix.

    B : array_like, shape `(rank + oversample, p)`.
        Smaller matrix.

    References
    ----------
    N. Halko, P. Martinsson, and J. Tropp.
    "Finding structure with randomness: probabilistic
    algorithms for constructing approximate matrix
    decompositions" (2009).
    (available at `arXiv <http://arxiv.org/abs/0909.4061>`_).

    S. Voronin and P.Martinsson.
    "RSVDPACK: Subroutines for computing partial singular value
    decompositions via randomized sampling on single core, multi core,
    and GPU architectures" (2015).
    (available at `arXiv <http://arxiv.org/abs/1502.05366>`_).

    """
    # TODO: adapt for complex-valued data
    if n_blocks > 1:
        if isinstance(A, DaskArray):
            raise ValueError(
                f"For delayed-dask computation n_blocks must be 1, but found n_blocks={n_blocks}."
            )
        n, _ = A.shape

        # index sets
        row_sets = np.array_split(range(n), n_blocks)

        Q_block = []
        K = []

        nblock = 1
        for rows in row_sets:
            # converts A to array, raise ValueError if A has inf or nan
            Qtemp, Ktemp = _compute_rqb(
                A[rows, :],
                rank=rank,
                oversample=oversample,
                n_subspace=n_subspace,
                random_state=random_state,
            )

            Q_block.append(Qtemp)
            K.append(Ktemp)
            nblock += 1

        Q_small, B = _compute_rqb(
            np.concatenate(K, axis=0),
            rank=rank,
            oversample=oversample,
            n_subspace=n_subspace,
            random_state=random_state,
        )

        Q_small = np.vsplit(Q_small, n_blocks)

        Q = [Q_block[i].dot(Q_small[i]) for i in range(n_blocks)]
        Q = np.concatenate(Q, axis=0)

    else:
        Q, B = _compute_rqb(
            A,
            rank=rank,
            oversample=oversample,
            n_subspace=n_subspace,
            random_state=random_state,
        )

    return Q, B


def soft_l0(arr, thresh):
    """Soft threshold operator for l0 regularization.

    Parameters
    ----------
    arr : array_like
        Input array.
    thresh : float
        Threshold value.

    Returns
    -------
    array_like
        Thresholded array.

    """
    # TODO: adapt for complex-valued data
    idx = arr**2 < 2 * thresh
    arr[idx] = 0
    return arr


def soft_l1(arr, thresh):
    """Soft threshold operator for l1 regularization.

    Parameters
    ----------
    arr : array_like
        Input array.
    thresh : float
        Threshold value.

    Returns
    -------
    array_like
        Thresholded array.

    """
    # TODO: adapt for complex-valued data
    return np.sign(arr) * np.maximum(np.abs(arr) - thresh, 0)


def compute_residual(X, B, A):
    # TODO: adapt for complex-valued data
    return X - X.dot(B).dot(A.T)


def compute_spca(
    X,
    n_components=None,
    alpha=0.1,
    beta=1e-5,
    gamma=0.1,
    robust=False,
    regularizer="l1",
    max_iter=1e3,
    tol=1e-5,
    compute=False,
):
    r"""Sparse Principal Component Analysis (SPCA).

    Given a mean centered rectangular matrix `A` with shape `(m, n)`, SPCA
    computes a set of sparse components that can optimally reconstruct the
    input data.  The amount of sparseness is controllable by the coefficient
    of the L1 penalty, given by the parameter alpha. In addition, some ridge
    shrinkage can be applied in order to improve conditioning.


    Parameters
    ----------
    X : array_like, shape `(n, p)`.
        Input array.

    n_components : integer, `n_components << min{m,n}`.
        Target rank, i.e., number of sparse components to be computed.

    alpha : float, (default ``alpha = 0.1``).
        Sparsity controlling parameter. Higher values lead to sparser components.

    beta : float, (default ``beta = 1e-5``).
        Amount of ridge shrinkage to apply in order to improve conditionin.

    regularizer : string {'l0', 'l1'}.
        Type of sparsity-inducing regularizer. The l1 norm (also known as LASSO)
        leads to softhreshold operator (default).  The l0 norm is implemented
        via a hardthreshold operator.

    robust : bool ``{'True', 'False'}``, optional (default ``False``).
        Use a robust algorithm to compute the sparse PCA.

    max_iter : integer, (default ``max_iter = 500``).
        Maximum number of iterations to perform before exiting.

    tol : float, (default ``tol = 1e-5``).
        Stopping tolerance for reconstruction error.

    compute: bool
        Whether to eagerly compute the rotation matrix. If ``True``, then
        the rotation algorithm will check for convergence at each iteration
        and stop once reached. If ``False``, then the rotation algorithm
        can build the matrix lazily, but it will not check for convergence
        and will run all of ``max_iter``.

    Returns
    -------
    B:  array_like, `(p, n_components)`.
        Sparse components extracted from the data.

    A : array_like, `(p, n_components)`.
        Orthogonal components extracted from the data.

    eigvals : array_like, `(n_components)`.
        Eigenvalues correspnding to the extracted components.

    Notes
    -----
    Variable Projection for PCA solves the following optimization problem:
    minimize :math:`1/2 \| X - X B A^T \|^2 + \alpha \|B\|_1 + 1/2 \beta \|B\|^2`
    """
    # TODO: adapt for complex-valued data

    if isinstance(X, DaskArray):

        def svd_algorithm(X):
            return dask.array.linalg.svd(X)
    else:
        # TODO: can we pass the arguments without using functions?
        # linalg.svd(X, full_matrices=False, overwrite_a=False)
        # perhaps we can use Decomposer class to handle this
        def svd_algorithm(X):
            return scipy.linalg.svd(X, full_matrices=False, overwrite_a=False)

    if regularizer == "l1":
        regularizer_func = soft_l1
    elif regularizer == "l0":
        if robust:
            raise NotImplementedError(
                "l0 regularization is not supported for " "robust sparse pca"
            )
        regularizer_func = soft_l0
    else:
        raise ValueError(
            'regularizer must be one of ("l1", "l0"), not ' "%s." % regularizer
        )

    m, n = X.shape
    if n_components is not None:
        if n_components > n:
            raise ValueError(
                "n_components must be less than the number " "of columns of X (%d)" % n
            )
    else:
        n_components = n

    # Initialization of Variable Projection Solver
    U, D, Vt = svd_algorithm(X)
    Dmax = D[0]  # l2 norm

    A = Vt[:n_components].T
    B = Vt[:n_components].T

    if robust:
        U = U[:, :n_components]
        Vt = Vt[:n_components]
        S = np.zeros_like(X)
    else:
        # compute outside the loop
        VD = Vt.T * D
        VD2 = Vt.T * D**2

    # Set Tuning Parameters
    alpha *= Dmax**2
    beta *= Dmax**2
    nu = 1.0 / (Dmax**2 + beta)
    kappa = nu * alpha

    obj = []  # values of objective function
    n_iter = 0

    #   Apply Variable Projection Solver
    while max_iter > n_iter:
        # Update A:
        # X'XB = UDV'
        # Compute X'XB via SVD of X
        if robust:
            XS = X - S
            XB = X.dot(B)
            Z = (XS).T.dot(XB)
        else:
            Z = VD2.dot(Vt.dot(B))

        Utilde, Dtilde, Vttilde = svd_algorithm(Z)
        A = Utilde.dot(Vttilde)

        # Proximal Gradient Descent to Update B
        if robust:
            R = XS - XB.dot(A.T)
            G = X.T.dot(R.dot(A)) - beta * B
        else:
            G = VD2.dot(Vt.dot(A - B)) - beta * B

        B = regularizer_func(B + nu * G, kappa)

        if robust:
            R = compute_residual(X, B, A)
            S = soft_l1(R, gamma)
            R -= S
        else:
            R = compute_residual(VD.T, B, A)

        objective = (
            0.5 * np.sum(R**2) + alpha * np.sum(np.abs(B)) + 0.5 * beta * np.sum(B**2)
        )
        if robust:
            objective += gamma * np.sum(np.abs(S))

        obj.append(objective)

        # Break if obj is not improving anymore
        if compute and n_iter > 0 and abs(obj[-2] - obj[-1]) / obj[-1] < tol:
            break

        # Next iter
        n_iter += 1

    eigen_values = Dtilde / (m - 1)

    return B, A, eigen_values


def compute_rspca(
    X,
    n_components,
    alpha=0.1,
    beta=0.1,
    max_iter=1e3,
    regularizer="l1",
    tol=1e-5,
    oversample=50,
    n_subspace=2,
    n_blocks=1,
    robust=False,
    random_state=None,
    compute=False,
):
    r"""Randomized Sparse Principal Component Analysis (rSPCA).

    Given a mean centered rectangular matrix `A` with shape `(n, p)`, SPCA
    computes a set of sparse components that can optimally reconstruct the
    input data.  The amount of sparseness is controllable by the coefficient
    of the L1 penalty, given by the parameter alpha. In addition, some ridge
    shrinkage can be applied in order to improve conditioning.

    This algorithm uses randomized methods for linear algebra to accelerate
    the computations.

    The quality of the approximation can be controlled via the oversampling
    parameter `oversample` and `n_subspace` which specifies the number of
    subspace iterations.


    Parameters
    ----------
    X : array_like, shape `(n, p)`.
        Real nonnegative input matrix.

    n_components : integer, `n_components << min{m,n}`.
        Target rank, i.e., number of sparse components to be computed.

    alpha : float, (default ``alpha = 0.1``).
        Sparsity controlling parameter. Higher values lead to sparser components.

    beta : float, (default ``beta = 0.1``).
        Amount of ridge shrinkage to apply in order to improve conditionin.

    regularizer : string {'l0', 'l1'}.
        Type of sparsity-inducing regularizer. The l1 norm (also known as LASSO)
        leads to softhreshold operator (default).  The l0 norm is implemented
        via a hardthreshold operator.

    max_iter : integer, (default ``max_iter = 500``).
        Maximum number of iterations to perform before exiting.

    tol : float, (default ``tol = 1e-5``).
        Stopping tolerance for reconstruction error.

    oversample : integer, optional (default: 10)
        Controls the oversampling of column space. Increasing this parameter
        may improve numerical accuracy.

    n_subspace : integer, default: 2.
        Parameter to control number of subspace iterations. Increasing this
        parameter may improve numerical accuracy.

    n_blocks : integer, default: 2.
        Paramter to control in how many blocks of columns the input matrix
        should be split. A larger number requires less fast memory, while it
        leads to a higher computational time.

    random_state : integer, RandomState instance or None, optional (default ``None``)
        If integer, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used by np.random.

    Returns
    -------
    B:  array_like, `(p, n_components)`.
        Sparse components extracted from the data.

    A : array_like, `(p, n_components)`.
        Orthogonal components extracted from the data.

    eigvals : array_like, `(n_components)`.
        Eigenvalues correspnding to the extracted components.

    S : array_like, `(n, p)`.
        Sparse component which captures grossly corrupted entries in the data
        matrix. Returned only if `robust == True`

    obj : array_like, `(n_iter)`.
        Objective value at the i-th iteration.

    Notes
    -----
    Variable Projection for SPCA solves the following optimization problem:
    minimize :math:`1/2 \| X - X B A^T \|^2 + \alpha \|B\|_1 + 1/2 \beta \|B\|^2`
    """

    # TODO: adapt for complex-valued data

    # Shape of data matrix
    m = X.shape[0]

    # Compute QB decomposition
    _, Xcompressed = compute_rqb(
        X,
        rank=n_components,
        oversample=oversample,
        n_subspace=n_subspace,
        n_blocks=n_blocks,
        random_state=random_state,
    )

    # Compute Sparse PCA
    B, A, eigen_values = compute_spca(
        Xcompressed,
        n_components=n_components,
        alpha=alpha,
        beta=beta,
        regularizer=regularizer,
        max_iter=max_iter,
        tol=tol,
        robust=robust,
        compute=compute,
    )
    # rescale eigen values
    eigen_values *= (n_components + oversample - 1) / (m - 1)

    return B, A, eigen_values



import warnings

import numpy as np
import xarray as xr
from typing_extensions import Self

from ..linalg.decomposer import Decomposer
from ..utils.data_types import DataArray, DataObject
from ..utils.hilbert_transform import hilbert_transform
from ..utils.xarray_utils import total_variance as compute_total_variance
from .base_model_single_set import BaseModelSingleSet


class EOF(BaseModelSingleSet):
    """EOF analysis.

    Empirical Orthogonal Functions (EOF) analysis, more commonly known
    as Principal Component Analysis (PCA).

    Parameters
    ----------
    n_modes: int, default=10
        Number of modes to calculate.
    center: bool, default=True
        Whether to center the input data.
    standardize: bool, default=False
        Whether to standardize the input data.
    use_coslat: bool, default=False
        Whether to use cosine of latitude for scaling.
    sample_name: str, default="sample"
        Name of the sample dimension.
    feature_name: str, default="feature"
        Name of the feature dimension.
    compute : bool, default=True
        Whether to compute elements of the model eagerly, or to defer computation.
        If True, four pieces of the fit will be computed sequentially: 1) the
        preprocessor scaler, 2) optional NaN checks, 3) SVD decomposition, 4) scores
        and components.
    random_state : int, optional
        Seed for the random number generator.
    solver: {"auto", "full", "randomized"}, default="auto"
        Solver to use for the SVD computation.
    solver_kwargs: dict, default={}
        Additional keyword arguments to be passed to the SVD solver.

    Examples
    --------
    >>> model = xe.single.EOF(n_modes=5)
    >>> model.fit(X)
    >>> scores = model.scores()

    """

    def __init__(
        self,
        n_modes: int = 2,
        center: bool = True,
        standardize: bool = False,
        use_coslat: bool = False,
        check_nans=True,
        sample_name: str = "sample",
        feature_name: str = "feature",
        compute: bool = True,
        random_state: int | None = None,
        solver: str = "auto",
        solver_kwargs: dict = {},
        **kwargs,
    ):
        super().__init__(
            n_modes=n_modes,
            center=center,
            standardize=standardize,
            use_coslat=use_coslat,
            check_nans=check_nans,
            sample_name=sample_name,
            feature_name=feature_name,
            compute=compute,
            random_state=random_state,
            solver=solver,
            solver_kwargs=solver_kwargs,
            **kwargs,
        )
        self.attrs.update({"model": "EOF analysis"})

    def _fit_algorithm(self, X: DataArray) -> Self:
        sample_name = self.sample_name
        feature_name = self.feature_name

        # Augment the data
        X = self._augment_data(X)

        # Compute the total variance
        total_variance = compute_total_variance(X, dim=sample_name)

        # Decompose the data
        decomposer = Decomposer(**self._decomposer_kwargs)
        decomposer.fit(X, dims=(sample_name, feature_name))

        singular_values = decomposer.s_
        components = decomposer.V_
        scores = decomposer.U_ * decomposer.s_
        scores.name = "scores"

        # Compute the explained variance per mode
        n_samples = X.coords[self.sample_name].size
        exp_var = singular_values**2 / (n_samples - 1)
        exp_var.name = "explained_variance"

        # Store the results
        self.data.add(X, "input_data", allow_compute=False)
        self.data.add(components, "components")
        self.data.add(scores, "scores")
        self.data.add(singular_values, "norms")
        self.data.add(exp_var, "explained_variance")
        self.data.add(total_variance, "total_variance")

        self.data.set_attrs(self.attrs)
        return self

    def _augment_data(self, X: DataArray) -> DataArray:
        return X

    def _transform_algorithm(self, X: DataObject) -> DataArray:
        feature_name = self.preprocessor.feature_name

        components = self.data["components"]

        # Project the data
        projections = xr.dot(X, components, dims=feature_name)
        projections.name = "scores"

        return projections

    def _inverse_transform_algorithm(self, scores: DataArray) -> DataArray:
        """Reconstruct the original data from transformed data.

        Parameters
        ----------
        scores: DataArray
            Transformed data to be reconstructed. This could be a subset
            of the `scores` data of a fitted model, or unseen data. Must
            have a 'mode' dimension.

        Returns
        -------
        data: DataArray | Dataset | list[DataArray]
            Reconstructed data.

        """
        # Reconstruct the data
        comps = self.data["components"].sel(mode=scores.mode)

        reconstructed_data = xr.dot(comps.conj(), scores, dims="mode")
        reconstructed_data.name = "reconstructed_data"

        return reconstructed_data

    def components(self, normalized: bool = True) -> DataObject:
        """Return the components.

        The components are also refered to as eigenvectors, EOFs or loadings depending on the context.

        Returns
        -------
        components: DataArray | Dataset | list[DataArray]
            Components of the fitted model.

        """
        return super().components(normalized=normalized)

    def scores(self, normalized: bool = False) -> DataArray:
        """Return the (PC) scores.

        The scores in EOF anaylsis are the projection of the data matrix onto the
        eigenvectors of the covariance matrix (or correlation) matrix.
        Other names include the principal component (PC) scores or just PCs.

        Parameters
        ----------
        normalized : bool, default=True
            Whether to normalize the scores by the L2 norm (singular values).

        Returns
        -------
        components: DataArray | Dataset | list[DataArray]
            Scores of the fitted model.

        """
        return super().scores(normalized=normalized)

    def singular_values(self) -> DataArray:
        """Return the singular values of the Singular Value Decomposition.

        Returns
        -------
        singular_values: DataArray
            Singular values obtained from the SVD.

        """
        return self.data["norms"]

    def explained_variance(self) -> DataArray:
        """Return explained variance.

        The explained variance :math:`\\lambda_i` is the variance explained
        by each mode. It is defined as

        .. math::
            \\lambda_i = \\frac{\\sigma_i^2}{N-1}

        where :math:`\\sigma_i` is the singular value of the :math:`i`-th mode and :math:`N` is the number of samples.
        Equivalently, :math:`\\lambda_i` is the :math:`i`-th eigenvalue of the covariance matrix.

        Returns
        -------
        explained_variance: DataArray
            Explained variance.
        """
        return self.data["explained_variance"]

    def explained_variance_ratio(self) -> DataArray:
        """Return explained variance ratio.

        The explained variance ratio :math:`\\gamma_i` is the variance explained
        by each mode normalized by the total variance. It is defined as

        .. math::
            \\gamma_i = \\frac{\\lambda_i}{\\sum_{j=1}^M \\lambda_j}

        where :math:`\\lambda_i` is the explained variance of the :math:`i`-th mode and :math:`M` is the total number of modes.

        Returns
        -------
        explained_variance_ratio: DataArray
            Explained variance ratio.
        """
        exp_var_ratio = self.data["explained_variance"] / self.data["total_variance"]
        exp_var_ratio.attrs.update(self.data["explained_variance"].attrs)
        exp_var_ratio.name = "explained_variance_ratio"
        return exp_var_ratio


class ComplexEOF(EOF):
    """Complex EOF analysis.

    EOF analysis applied to a complex-valued field obtained from a pair of
    variables such as the zonal and meridional components, :math:`U` and
    :math:`V`, of the wind field. Complex EOF analysis then decomposes the
    dataset

    .. math::
        Z = U + iV

    into a set of complex-valued components (EOFs) and PC scores [1]_.

    Parameters
    ----------
    n_modes : int
        Number of modes to calculate.
    center: bool, default=True
        Whether to center the input data.
    standardize : bool
        Whether to standardize the input data.
    use_coslat : bool
        Whether to use cosine of latitude for scaling.
    sample_name: str, default="sample"
        Name of the sample dimension.
    feature_name: str, default="feature"
        Name of the feature dimension.
    compute : bool, default=True
        Whether to compute elements of the model eagerly, or to defer
        computation. If True, four pieces of the fit will be computed
        sequentially: 1) the preprocessor scaler, 2) optional NaN checks, 3) SVD
        decomposition, 4) scores and components.
    random_state : int, optional
        Seed for the random number generator.
    solver: {"auto", "full", "randomized"}, default="auto"
        Solver to use for the SVD computation.
    solver_kwargs: dict, default={}
        Additional keyword arguments to be passed to the SVD solver.

    References
    ----------
    .. [1] Storch, H. von & Zwiers, F. W. Statistical Analysis in Climate
        Research. (Cambridge University Press (Virtual Publishing), 2003).


    Examples
    --------

    With two DataArrays `u` and `v` representing the zonal and meridional
    components of the wind field, construct

    >>> X = u + 1j * v

    and fit the Complex EOF model:

    >>> model = ComplexEOF(n_modes=5, standardize=True)
    >>> model.fit(X, "time")

    """

    uses_complex = True

    def __init__(
        self,
        n_modes: int = 2,
        center: bool = True,
        standardize: bool = False,
        use_coslat: bool = False,
        check_nans: bool = True,
        sample_name: str = "sample",
        feature_name: str = "feature",
        compute: bool = True,
        random_state: int | None = None,
        solver: str = "auto",
        solver_kwargs: dict = {},
        **kwargs,
    ):
        super().__init__(
            n_modes=n_modes,
            center=center,
            standardize=standardize,
            use_coslat=use_coslat,
            check_nans=check_nans,
            sample_name=sample_name,
            feature_name=feature_name,
            compute=compute,
            random_state=random_state,
            solver=solver,
            solver_kwargs=solver_kwargs,
            **kwargs,
        )
        self.attrs.update({"model": "Complex EOF analysis"})

    def _fit_algorithm(self, X: DataArray) -> Self:
        if not np.iscomplexobj(X):
            warnings.warn(
                "Expected complex-valued data but found real-valued data. For Hilbert EOF analysis, use `HilbertEOF` model."
            )

        return super()._fit_algorithm(X)

    def components_amplitude(self, normalized=True) -> DataObject:
        """Return the amplitude of the (EOF) components.

        The amplitude of the components are defined as

        .. math::
            A_{ij} = |C_{ij}|

        where :math:`C_{ij}` is the :math:`i`-th entry of the :math:`j`-th component and
        :math:`|\\cdot|` denotes the absolute value.

        Parameters
        ----------
        normalized : bool, default=True
            Whether to normalize the components by the singular values

        Returns
        -------
        components_amplitude: DataArray | Dataset | list[DataArray]
            Amplitude of the components of the fitted model.

        """
        amplitudes = abs(self.data["components"])

        if not normalized:
            amplitudes = amplitudes * self.data["norms"]

        amplitudes.name = "components_amplitude"
        return self.preprocessor.inverse_transform_components(amplitudes)

    def components_phase(self) -> DataObject:
        """Return the phase of the (EOF) components.

        The phase of the components are defined as

        .. math::
            \\phi_{ij} = \\arg(C_{ij})

        where :math:`C_{ij}` is the :math:`i`-th entry of the :math:`j`-th component and
        :math:`\\arg(\\cdot)` denotes the argument of a complex number.

        Returns
        -------
        components_phase: DataArray | Dataset | list[DataArray]
            Phase of the components of the fitted model.

        """
        comps = self.data["components"]
        comp_phase = xr.apply_ufunc(np.angle, comps, dask="allowed", keep_attrs=True)
        comp_phase.name = "components_phase"
        return self.preprocessor.inverse_transform_components(comp_phase)

    def scores_amplitude(self, normalized=True) -> DataArray:
        """Return the amplitude of the (PC) scores.

        The amplitude of the scores are defined as

        .. math::
            A_{ij} = |S_{ij}|

        where :math:`S_{ij}` is the :math:`i`-th entry of the :math:`j`-th score and
        :math:`|\\cdot|` denotes the absolute value.

        Parameters
        ----------
        normalized : bool, default=True
            Whether to normalize the scores by the singular values.

        Returns
        -------
        scores_amplitude: DataArray | Dataset | list[DataArray]
            Amplitude of the scores of the fitted model.

        """
        scores = self.data["scores"].copy()
        if normalized:
            scores = scores / self.data["norms"]

        amplitudes = abs(scores)
        amplitudes.name = "scores_amplitude"
        return self.preprocessor.inverse_transform_scores(amplitudes)

    def scores_phase(self) -> DataArray:
        """Return the phase of the (PC) scores.

        The phase of the scores are defined as

        .. math::
            \\phi_{ij} = \\arg(S_{ij})

        where :math:`S_{ij}` is the :math:`i`-th entry of the :math:`j`-th score and
        :math:`\\arg(\\cdot)` denotes the argument of a complex number.

        Returns
        -------
        scores_phase: DataArray | Dataset | list[DataArray]
            Phase of the scores of the fitted model.

        """
        scores = self.data["scores"]
        phases = xr.apply_ufunc(np.angle, scores, dask="allowed", keep_attrs=True)
        phases.name = "scores_phase"
        return self.preprocessor.inverse_transform_scores(phases)


class HilbertEOF(ComplexEOF):
    """Hilbert EOF analysis.

    The Hilbert EOF analysis [1]_ [2]_ [3]_ [4]_ (also known as Hilbert EOF analysis) applies a Hilbert transform
    to the data before performing the standard EOF analysis.
    The Hilbert transform is applied to each feature of the data individually.

    An optional padding with exponentially decaying values can be applied prior to
    the Hilbert transform in order to mitigate the impact of spectral leakage.

    Parameters
    ----------
    n_modes : int
        Number of modes to calculate.
    padding : str, optional
        Specifies the method used for padding the data prior to applying the Hilbert
        transform. This can help to mitigate the effect of spectral leakage.
        Currently, only 'exp' for exponential padding is supported. Default is 'exp'.
    decay_factor : float, optional
        Specifies the decay factor used in the exponential padding. This parameter
        is only used if padding='exp'. The recommended value typically ranges between 0.05 to 0.2
        but ultimately depends on the variability in the data.
        A smaller value (e.g. 0.05) is recommended for
        data with high variability, while a larger value (e.g. 0.2) is recommended
        for data with low variability. Default is 0.2.
    center: bool, default=True
        Whether to center the input data.
    standardize : bool
        Whether to standardize the input data.
    use_coslat : bool
        Whether to use cosine of latitude for scaling.
    sample_name: str, default="sample"
        Name of the sample dimension.
    feature_name: str, default="feature"
        Name of the feature dimension.
    compute : bool, default=True
        Whether to compute elements of the model eagerly, or to defer computation.
        If True, four pieces of the fit will be computed sequentially: 1) the
        preprocessor scaler, 2) optional NaN checks, 3) SVD decomposition, 4) scores
        and components.
    random_state : int, optional
        Seed for the random number generator.
    solver: {"auto", "full", "randomized"}, default="auto"
        Solver to use for the SVD computation.
    solver_kwargs: dict, default={}
        Additional keyword arguments to be passed to the SVD solver.
    solver_kwargs : dict, optional
        Additional keyword arguments to be passed to the SVD solver.

    References
    ----------
    .. [1] Rasmusson, E. M., Arkin, P. A., Chen, W.-Y. & Jalickee, J. B. Biennial variations in surface temperature over the United States as revealed by singular decomposition. Monthly Weather Review 109, 587–598 (1981).
    .. [2] Barnett, T. P. Interaction of the Monsoon and Pacific Trade Wind System at Interannual Time Scales Part I: The Equatorial Zone. Monthly Weather Review 111, 756–773 (1983).
    .. [3] Horel, J. Complex Principal Component Analysis: Theory and Examples. J. Climate Appl. Meteor. 23, 1660–1673 (1984).
    .. [4] Hannachi, A., Jolliffe, I. & Stephenson, D. Empirical orthogonal functions and related techniques in atmospheric science: A review. International Journal of Climatology 27, 1119–1152 (2007).

    Examples
    --------
    >>> model = HilbertEOF(n_modes=5, standardize=True)
    >>> model.fit(X)

    """

    def __init__(
        self,
        n_modes: int = 2,
        padding: str = "exp",
        decay_factor: float = 0.2,
        center: bool = True,
        standardize: bool = False,
        use_coslat: bool = False,
        check_nans: bool = True,
        sample_name: str = "sample",
        feature_name: str = "feature",
        compute: bool = True,
        random_state: int | None = None,
        solver: str = "auto",
        solver_kwargs: dict = {},
        **kwargs,
    ):
        super().__init__(
            n_modes=n_modes,
            center=center,
            standardize=standardize,
            use_coslat=use_coslat,
            check_nans=check_nans,
            sample_name=sample_name,
            feature_name=feature_name,
            compute=compute,
            random_state=random_state,
            solver=solver,
            solver_kwargs=solver_kwargs,
            **kwargs,
        )
        self.attrs.update({"model": "Hilbert EOF analysis"})
        self._params.update({"padding": padding, "decay_factor": decay_factor})

    def _augment_data(self, X: DataArray) -> DataArray:
        # Apply hilbert transform:
        padding = self._params["padding"]
        decay_factor = self._params["decay_factor"]
        return hilbert_transform(
            X,
            dims=(self.sample_name, self.feature_name),
            padding=padding,
            decay_factor=decay_factor,
        )

    def _fit_algorithm(self, X: DataArray) -> Self:
        EOF._fit_algorithm(self, X)
        return self

    def _transform_algorithm(self, X: DataArray) -> DataArray:
        raise NotImplementedError("Hilbert EOF does not support transform method.")

    def _inverse_transform_algorithm(self, scores: DataArray) -> DataArray:
        Xrec = super()._inverse_transform_algorithm(scores)
        # Enforce real output
        return Xrec.real



from .concatenator import Concatenator
from .dimension_renamer import DimensionRenamer
from .multi_index_converter import MultiIndexConverter
from .pca import PCA
from .preprocessor import Preprocessor
from .sanitizer import Sanitizer
from .scaler import Scaler
from .stacker import Stacker
from .whitener import Whitener

__all__ = [
    "Concatenator",
    "DimensionRenamer",
    "MultiIndexConverter",
    "Preprocessor",
    "Sanitizer",
    "Scaler",
    "Stacker",
    "Whitener",
    "PCA",
]



import numpy as np
import xarray as xr
from dask.base import compute as dask_compute

from ..utils.data_types import DataArray
from ._numpy import _SVD


class SVD:
    def __init__(
        self,
        n_modes: int | float | str,
        is_complex: bool | str = "auto",
        init_rank_reduction: float = 0.3,
        flip_signs: bool = True,
        solver: str = "auto",
        compute: bool = True,
        random_state: np.random.Generator | int | None = None,
        solver_kwargs: dict = {},
        sample_name: str = "sample",
        feature_name: str = "feature",
    ):
        self.n_modes = n_modes
        self.is_complex = is_complex
        self.init_rank_reduction = init_rank_reduction
        self.flip_signs = flip_signs
        self.solver = solver
        self.random_state = random_state
        self.solver_kwargs = solver_kwargs
        self.compute_svd = compute

        self.sample_name = sample_name
        self.feature_name = feature_name

    def fit_transform(self, X: DataArray) -> tuple[DataArray, DataArray, DataArray]:
        """Decomposes the data object.

        Parameters
        ----------
        X : DataArray
            A 2-dimensional data object to be decomposed.

        Returns
        -------
        U : DataArray
            The left singular vectors of the decomposition.
        s : DataArray
            The singular values of the decomposition.
        V : DataArray
            The right singular vectors of the decomposition.

        """
        svd = _SVD(
            n_modes=self.n_modes,
            init_rank_reduction=self.init_rank_reduction,
            flip_signs=self.flip_signs,
            solver=self.solver,
            random_state=self.random_state,
            is_complex=self.is_complex,
            **self.solver_kwargs,
        )
        U, s, V = xr.apply_ufunc(
            svd.fit_transform,
            X,
            input_core_dims=[[self.sample_name, self.feature_name]],
            output_core_dims=[
                [self.sample_name, "mode"],
                ["mode"],
                [self.feature_name, "mode"],
            ],
            dask="allowed",
        )
        mode_coords = np.arange(1, U.mode.size + 1)
        s = s.assign_coords(mode=mode_coords)
        U = U.assign_coords(mode=mode_coords)
        V = V.assign_coords(mode=mode_coords)

        if self.compute_svd:
            U, s, V = dask_compute(U, s, V)

        return U, s, V



from abc import ABC, abstractmethod

import pandas as pd
import xarray as xr
from sklearn.base import BaseEstimator, TransformerMixin
from typing_extensions import Self

from ..utils.data_types import Data, DataArray, DataSet, Dims


class Transformer(BaseEstimator, TransformerMixin, ABC):
    """
    Abstract base class to transform an xarray DataArray/Dataset.

    """

    def __init__(
        self,
        sample_name: str = "sample",
        feature_name: str = "feature",
    ):
        self.sample_name = sample_name
        self.feature_name = feature_name

    @abstractmethod
    def get_serialization_attrs(self) -> dict:
        """Return a dictionary containing the attributes that need to be serialized
        as part of a saved transformer.

        There are limitations on the types of attributes that can be serialized.
        Most simple types (e.g. int, float, str, bool, None) can be, as well as
        DataArrays and dicts of DataArrays. Other nested types (e.g. lists of
        DataArrays) will likely fail.

        """
        return dict()

    @abstractmethod
    def fit(
        self,
        X: Data,
        sample_dims: Dims | None = None,
        feature_dims: Dims | None = None,
        **kwargs,
    ) -> Self:
        """Fit transformer to data.

        Parameters:
        -------------
        X: xr.DataArray | xr.Dataset
            Input data.
        sample_dims: Sequence[Hashable], optional
            Sample dimensions.
        feature_dims: Sequence[Hashable], optional
            Feature dimensions.
        """
        pass

    @abstractmethod
    def transform(self, X: Data) -> Data:
        return X

    def fit_transform(
        self,
        X: Data,
        sample_dims: Dims | None = None,
        feature_dims: Dims | None = None,
        **kwargs,
    ) -> Data:
        return self.fit(X, sample_dims, feature_dims, **kwargs).transform(X)

    @abstractmethod
    def inverse_transform_data(self, X: Data) -> Data:
        return X

    @abstractmethod
    def inverse_transform_components(self, X: Data) -> Data:
        return X

    @abstractmethod
    def inverse_transform_scores(self, X: DataArray) -> DataArray:
        return X

    @abstractmethod
    def inverse_transform_scores_unseen(self, X: DataArray) -> DataArray:
        return X

    def _serialize_data(self, key: str, data: Data) -> DataSet:
        multiindexes = {}
        name_map = None
        if isinstance(data, xr.Dataset):
            # Keep Dataset as is
            ds = data
        else:
            # Convert DataArray to Dataset
            coords = {}
            data_vars = {}
            if data.name in data.coords:
                # Convert a coord-like DataArray to Dataset and note multiindexes
                if isinstance(data.to_index(), pd.MultiIndex):
                    multiindexes[data.name] = [n for n in data.to_index().names]
                coords[data.name] = data
            else:
                # Make sure the DataArray has some name so we can create a string mapping
                if data.name is None:
                    data.name = key
                data_vars[data.name] = data
            ds = xr.Dataset(data_vars=data_vars, coords=coords)
            name_map = data.name

        # Drop multiindexes and record for later
        ds = ds.reset_index(list(multiindexes.keys()))
        ds.attrs["multiindexes"] = multiindexes
        ds.attrs["name_map"] = {key: name_map}

        return ds

    def serialize(self) -> xr.DataTree:
        """Serialize a transformer to a DataTree."""
        return self._serialize()

    def _serialize(self) -> xr.DataTree:
        """Serialize a transformer to a DataTree. Use an internal
        method so we can override the public one in subclasesses but
        still use this."""
        dt = xr.DataTree()
        params = self.get_params()
        attrs = self.get_serialization_attrs()

        # Set initialization params as tree level attrs
        dt.attrs["params"] = params

        # Serialize each transformer attribute
        for key, attr in attrs.items():
            if isinstance(attr, (xr.DataArray, xr.Dataset)):
                # attach data to data_vars or coords
                ds = self._serialize_data(key, attr)
                dt[key] = xr.DataTree(ds, name=key)
                dt.attrs[key] = "_is_node"
            elif isinstance(attr, dict) and any(
                [isinstance(val, xr.DataArray) for val in attr.values()]
            ):
                # attach dict of data as branching tree
                dt_attr = xr.DataTree()
                for k, v in attr.items():
                    ds = self._serialize_data(k, v)
                    dt_attr[k] = xr.DataTree(ds, name=k)
                dt[key] = dt_attr
                dt.attrs[key] = "_is_tree"
            else:
                # attach simple types as dataset attrs
                dt.attrs[key] = attr

        return dt

    def _deserialize_data_node(self, key: str, dt: xr.DataTree) -> DataArray:
        # Rebuild multiindexes
        dt.dataset = dt.dataset.set_index(dt.attrs.get("multiindexes", {}))
        # Extract the DataArray or coord from the Dataset
        data_key = dt.attrs["name_map"][key]
        if data_key is not None:
            return dt[data_key]
        else:
            return dt.dataset

    @classmethod
    def deserialize(cls, dt: xr.DataTree) -> Self:
        """Deserialize a saved transformer from a DataTree."""
        return cls._deserialize(dt)

    @classmethod
    def _deserialize(cls, dt: xr.DataTree) -> Self:
        """Deserialize a saved transformer from a xr.DataTree. Use an internal
        method so we can override the public one in subclasesses but
        still use this."""
        # Create the object from params
        transformer = cls(**dt.attrs["params"])

        # Set attributes
        for key, attr in dt.attrs.items():
            if key == "params":
                continue
            elif attr == "_is_node":
                data = transformer._deserialize_data_node(key, dt[key])
                setattr(transformer, key, data)
            elif attr == "_is_tree":
                data = {}
                for k, v in dt[key].items():
                    data[k] = transformer._deserialize_data_node(k, dt[key][k])
                setattr(transformer, key, data)
            else:
                setattr(transformer, key, attr)

        return transformer



"""
Significance testing of EOF analysis via bootstrap
===================================================

Test the significance of individual modes and obtain confidence intervals
for both EOFs and PCs.
"""

# Load packages and data:
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
from cartopy.crs import Orthographic, PlateCarree
from matplotlib.gridspec import GridSpec

import xeofs as xe

# %%

t2m = xr.tutorial.load_dataset("air_temperature")["air"]

# %%
# Perform EOF analysis

model = xe.single.EOF(n_modes=5, standardize=False)
model.fit(t2m, dim="time")
expvar = model.explained_variance_ratio()
components = model.components()
scores = model.scores()


# %%
# Perform bootstrapping of the model to identy the number of significant modes.
# We perform 50 bootstraps.
# Note - if computationallly feasible - you typically want to choose higher
# numbers of bootstraps e.g. 1000.

n_boot = 50

bs = xe.validation.EOFBootstrapper(n_bootstraps=n_boot)
bs.fit(model)
bs_expvar = bs.explained_variance()
ci_expvar = bs_expvar.quantile([0.025, 0.975], "n")  # 95% confidence intervals

q025 = ci_expvar.sel(quantile=0.025)
q975 = ci_expvar.sel(quantile=0.975)

is_significant = q025 - q975.shift({"mode": -1}) > 0
n_significant_modes = (
    is_significant.where(is_significant is True).cumsum(skipna=False).max().fillna(0)
)
print("{:} modes are significant at alpha=0.05".format(n_significant_modes.values))

# %%
# The bootstrapping procedure identifies 3 significant modes. We can also
# compute the 95 % confidence intervals of the EOFs/PCs and mask out
# insignificant elements of the obtained EOFs.

ci_components = bs.components().quantile([0.025, 0.975], "n")
ci_scores = bs.scores().quantile([0.025, 0.975], "n")

is_sig_comps = np.sign(ci_components).prod("quantile") > 0


# %%
# Summarize the results in a figure.


lons, lats = np.meshgrid(is_sig_comps.lon.values, is_sig_comps.lat.values)
proj = Orthographic(central_latitude=30, central_longitude=-80)
kwargs = {"cmap": "RdBu", "vmin": -0.05, "vmax": 0.05, "transform": PlateCarree()}

fig = plt.figure(figsize=(10, 16))
gs = GridSpec(5, 2)
ax1 = [fig.add_subplot(gs[i, 0], projection=proj) for i in range(5)]
ax2 = [fig.add_subplot(gs[i, 1]) for i in range(5)]

for i, (a1, a2) in enumerate(zip(ax1, ax2)):
    a1.coastlines(color=".5")
    components.isel(mode=i).plot(ax=a1, **kwargs)
    a1.scatter(
        lons,
        lats,
        is_sig_comps.isel(mode=i).values * 0.5,
        color="k",
        alpha=0.5,
        transform=PlateCarree(),
    )
    ci_scores.isel(mode=i, quantile=0).plot(ax=a2, color=".3", lw=".5", label="2.5%")
    ci_scores.isel(mode=i, quantile=1).plot(ax=a2, color=".3", lw=".5", label="97.5%")
    scores.isel(mode=i).plot(ax=a2, lw=".5", alpha=0.5, label="PC")
    a2.legend(loc=2)

plt.tight_layout()
plt.savefig("bootstrap.jpg")



from datetime import datetime

import numpy as np
import xarray as xr
from typing_extensions import Self

from .._version import __version__
from ..data_container import DataContainer
from ..linalg.rotation import promax
from ..preprocessing import Preprocessor
from ..utils.data_types import DataArray
from ..utils.xarray_utils import argsort_dask, get_deterministic_sign_multiplier
from .eof import EOF, ComplexEOF, HilbertEOF


class EOFRotator(EOF):
    """Rotate a solution obtained from ``xe.single.EOF``.

    Rotated EOF analysis (e.g. [1]_) is a variation of standard EOF analysis, which uses a rotation technique
    (Varimax or Promax) on the extracted modes to maximize the variance explained by
    individual modes. This rotation spreads the explained variance more evenly among
    the modes, making them easier to interpret by minimizing the number of significant
    loadings on each mode.

    Parameters
    ----------
    n_modes : int, default=2
        Specify the number of modes to be rotated.
    power : int, default=1
        Set the power for the Promax rotation. A ``power`` value of 1 results
        in a Varimax rotation.
    max_iter : int or None, default=None
        Determine the maximum number of iterations for the computation of the
        rotation matrix. If not specified, defaults to 1000 if ``compute=True``
        and 100 if ``compute=False``, since we can't terminate a lazy computation
        based using ``rtol``.
    rtol : float, default=1e-8
        Define the relative tolerance required to achieve convergence and
        terminate the iterative process.
    compute : bool, default=True
        Whether to compute the rotation immediately.

    References
    ----------
    .. [1] Richman, M.B., 1986. Rotation of principal components. Journal of Climatology 6, 293–335. https://doi.org/10.1002/joc.3370060305

    Examples
    --------
    >>> model = xe.single.EOF(n_modes=10)
    >>> model.fit(X, "time")
    >>> rotator = xe.single.EOFRotator(n_modes=10)
    >>> rotator.fit(model)
    >>> rotator.components()

    """

    def __init__(
        self,
        n_modes: int = 2,
        power: int = 1,
        max_iter: int | None = None,
        rtol: float = 1e-8,
        compute: bool = True,
    ):
        if max_iter is None:
            max_iter = 1000 if compute else 100

        # Define model parameters
        self._params = {
            "n_modes": n_modes,
            "power": power,
            "max_iter": max_iter,
            "rtol": rtol,
            "compute": compute,
        }

        # Define analysis-relevant meta data
        self.attrs = {"model": "Rotated EOF analysis"}
        self.attrs.update(self._params)
        self.attrs.update(
            {
                "software": "xeofs",
                "version": __version__,
                "date": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            }
        )

        # Attach empty objects
        self.preprocessor = Preprocessor()
        self.data = DataContainer()
        self.model_data = DataContainer()

        self.sorted = False

    def get_serialization_attrs(self) -> dict:
        return dict(
            data=self.data,
            preprocessor=self.preprocessor,
            model_data=self.model_data,
            sorted=self.sorted,
        )

    def fit(self, model) -> Self:
        """Rotate the solution obtained from ``xe.single.EOF``.

        Parameters
        ----------
        model : ``xe.single.EOF``
            The EOF model to be rotated.

        """
        self._fit_algorithm(model)

        if self._params["compute"]:
            self.compute()

        return self

    def _fit_algorithm(self, model) -> Self:
        self.preprocessor = model.preprocessor
        self.sample_name = model.sample_name
        self.feature_name = model.feature_name
        self.sorted = False

        n_modes = self._params.get("n_modes")
        power = self._params.get("power")
        max_iter = self._params.get("max_iter")
        rtol = self._params.get("rtol")

        # Select modes to rotate
        components = model.data["components"].sel(mode=slice(1, n_modes))
        expvar = model.explained_variance().sel(mode=slice(1, n_modes))

        # Rotate loadings
        loadings = components * np.sqrt(expvar)
        promax_kwargs = {"power": power, "max_iter": max_iter, "rtol": rtol}
        rot_loadings, rot_matrix, phi_matrix = promax(
            loadings,
            feature_dim=self.feature_name,
            compute=self._params["compute"],
            **promax_kwargs,
        )

        # Assign coordinates to the rotation/correlation matrices
        rot_matrix = rot_matrix.assign_coords(
            mode_m=np.arange(1, rot_matrix.mode_m.size + 1),
            mode_n=np.arange(1, rot_matrix.mode_n.size + 1),
        )
        phi_matrix = phi_matrix.assign_coords(
            mode_m=np.arange(1, phi_matrix.mode_m.size + 1),
            mode_n=np.arange(1, phi_matrix.mode_n.size + 1),
        )

        # Reorder according to variance
        expvar = (abs(rot_loadings) ** 2).sum(self.feature_name)
        idx_modes_sorted = argsort_dask(expvar, "mode")[::-1]
        idx_modes_sorted.coords.update(expvar.coords)

        # Normalize loadings
        rot_components = rot_loadings / np.sqrt(expvar)

        # Compute "pseudo" norms
        n_samples = model.data["input_data"].coords[self.sample_name].size
        norms = (expvar * (n_samples - 1)) ** 0.5
        norms.name = "singular_values"

        # Get unrotated, normalized scores
        svals = model.data["norms"].sel(mode=slice(1, n_modes))
        scores = model.data["scores"].sel(mode=slice(1, n_modes))
        scores = scores / svals
        # Rotate scores
        RinvT = self._compute_rot_mat_inv_trans(
            rot_matrix, input_dims=("mode_m", "mode_n")
        )
        # Rename dimension mode to ensure that dot product has dimensions (sample x mode) as output
        scores = scores.rename({"mode": "mode_m"})
        RinvT = RinvT.rename({"mode_n": "mode"})
        scores = xr.dot(scores, RinvT, dims="mode_m")

        # Scale scores by "pseudo" norms
        scores = scores * norms

        # Ensure consistent signs for deterministic output
        modes_sign = get_deterministic_sign_multiplier(
            rot_components, self.feature_name
        )
        rot_components = rot_components * modes_sign
        scores = scores * modes_sign

        # Store the results
        self.model_data.add(model.data["norms"], "singular_values")
        self.model_data.add(model.data["components"], "components")

        # Assigning input data to the Rotator object allows us to inherit some functionalities from the original model
        self.data.add(model.data["input_data"], "input_data", allow_compute=False)
        self.data.add(rot_components, "components")
        self.data.add(scores, "scores")
        self.data.add(norms, "norms")
        self.data.add(expvar, "explained_variance")
        self.data.add(model.data["total_variance"], "total_variance")
        self.data.add(idx_modes_sorted, "idx_modes_sorted")
        self.data.add(rot_matrix, "rotation_matrix")
        self.data.add(phi_matrix, "phi_matrix")
        self.data.add(modes_sign, "modes_sign")

        # Assign analysis-relevant meta data
        self.data.set_attrs(self.attrs)

        return self

    def _post_compute(self):
        """Leave sorting until after compute because it can't be done lazily."""
        self._sort_by_variance()

    def _sort_by_variance(self):
        """Re-sort the mode dimension of all data variables by variance explained."""
        if not self.sorted:
            for key in self.data.keys():
                if "mode" in self.data[key].dims and key != "idx_modes_sorted":
                    self.data[key] = (
                        self.data[key]
                        .isel(mode=self.data["idx_modes_sorted"].values)
                        .assign_coords(mode=self.data[key].mode)
                    )
        self.sorted = True

    def _transform_algorithm(self, X: DataArray) -> DataArray:
        n_modes = self._params["n_modes"]

        svals = self.model_data["singular_values"].sel(
            mode=slice(1, self._params["n_modes"])
        )
        pseudo_norms = self.data["norms"]
        # Select the (non-rotated) singular vectors of the first dataset
        components = self.model_data["components"].sel(mode=slice(1, n_modes))

        # Compute non-rotated scores by projecting the data onto non-rotated components
        projections = xr.dot(X, components) / svals
        projections.name = "scores"

        # Rotate the scores
        R = self.data["rotation_matrix"]
        RinvT = self._compute_rot_mat_inv_trans(R, input_dims=("mode_m", "mode_n"))
        projections = projections.rename({"mode": "mode_m"})
        RinvT = RinvT.rename({"mode_n": "mode"})
        projections = xr.dot(projections, RinvT, dims="mode_m")
        # Reorder according to variance
        # this must be done in one line: i) select modes according to their variance, ii) replace coords with modes from 1 ... n
        if self.sorted:
            projections = projections.isel(
                mode=self.data["idx_modes_sorted"].values
            ).assign_coords(mode=projections.mode)

        # Scale scores by "pseudo" norms
        projections = projections * pseudo_norms

        # Adapt the sign of the scores
        projections = projections * self.data["modes_sign"]

        return projections

    def fit_transform(self, model) -> DataArray:
        raise NotImplementedError(
            "The fit_transform method is not implemented for the EOFRotator class."
        )

    def _compute_rot_mat_inv_trans(self, rotation_matrix, input_dims) -> DataArray:
        """Compute the inverse transpose of the rotation matrix.

        For orthogonal rotations (e.g., Varimax), the inverse transpose is equivalent
        to the rotation matrix itself. For oblique rotations (e.g., Promax), the simplification
        does not hold.

        Returns
        -------
        rotation_matrix : xr.DataArray

        """
        if self._params["power"] > 1:
            # inverse matrix
            rotation_matrix = xr.apply_ufunc(
                np.linalg.inv,
                rotation_matrix,
                input_core_dims=[(input_dims)],
                output_core_dims=[(input_dims[::-1])],
                vectorize=False,
                dask="allowed",
            )
            # transpose matrix
            rotation_matrix = rotation_matrix.conj().transpose(*input_dims)
        return rotation_matrix


class ComplexEOFRotator(EOFRotator, ComplexEOF):
    """Rotate a solution obtained from ``xe.single.ComplexEOF``.

    Parameters
    ----------
    n_modes : int, default=2
        Specify the number of modes to be rotated.
    power : int, default=1
        Set the power for the Promax rotation. A ``power`` value of 1 results in
        a Varimax rotation.
    max_iter : int, default=1000
        Determine the maximum number of iterations for the computation of the
        rotation matrix.
    rtol : float, default=1e-8
        Define the relative tolerance required to achieve convergence and
        terminate the iterative process.
    compute: bool, default=True
        Whether to compute the rotation immediately.


    Examples
    --------
    >>> model = xe.single.ComplexEOF(n_modes=10)
    >>> model.fit(data)
    >>> rotator = xe.single.ComplexEOFRotator(n_modes=10)
    >>> rotator.fit(model)
    >>> rotator.components()


    """

    def __init__(
        self,
        n_modes: int = 2,
        power: int = 1,
        max_iter: int = 1000,
        rtol: float = 1e-8,
        compute: bool = True,
    ):
        super().__init__(
            n_modes=n_modes, power=power, max_iter=max_iter, rtol=rtol, compute=compute
        )
        self.attrs.update({"model": "Rotated Complex EOF analysis"})


class HilbertEOFRotator(EOFRotator, HilbertEOF):
    """Rotate a solution obtained from ``xe.single.HilbertEOF``.

    Hilbert Rotated EOF analysis [1]_ [2]_ [3]_ extends EOF analysis by incorporating both amplitude and phase information
    using a Hilbert transform prior to performing MCA and subsequent Varimax or Promax rotation.
    This adds a further layer of dimensionality to the analysis, allowing for a more nuanced interpretation
    of complex relationships within the data, particularly useful when analyzing oscillatory data.

    Parameters
    ----------
    n_modes : int, default=2
        Specify the number of modes to be rotated.
    power : int, default=1
        Set the power for the Promax rotation. A ``power`` value of 1 results
        in a Varimax rotation.
    max_iter : int, default=1000
        Determine the maximum number of iterations for the computation of the
        rotation matrix.
    rtol : float, default=1e-8
        Define the relative tolerance required to achieve convergence and
        terminate the iterative process.
    compute: bool, default=True
        Whether to compute the rotation immediately.

    References
    ----------
    .. [1] Horel, J., 1984. Complex Principal Component Analysis: Theory and Examples. J. Climate Appl. Meteor. 23, 1660–1673. https://doi.org/10.1175/1520-0450(1984)023<1660:CPCATA>2.0.CO;2
    .. [2] Richman, M.B., 1986. Rotation of principal components. Journal of Climatology 6, 293–335. https://doi.org/10.1002/joc.3370060305
    .. [3] Bloomfield, P., Davis, J.M., 1994. Orthogonal rotation of complex principal components. International Journal of Climatology 14, 759–775. https://doi.org/10.1002/joc.3370140706

    Examples
    --------
    >>> model = xe.single.HilbertEOF(n_modes=10)
    >>> model.fit(data)
    >>> rotator = xe.single.HilbertEOFRotator(n_modes=10)
    >>> rotator.fit(model)
    >>> rotator.components()


    """

    def __init__(
        self,
        n_modes: int = 2,
        power: int = 1,
        max_iter: int = 1000,
        rtol: float = 1e-8,
        compute: bool = True,
    ):
        super().__init__(
            n_modes=n_modes, power=power, max_iter=max_iter, rtol=rtol, compute=compute
        )
        self.attrs.update({"model": "Rotated Hilbert EOF analysis"})

    def _transform_algorithm(self, data: DataArray) -> DataArray:
        # Here we leverage the Method Resolution Order (MRO) to invoke the
        # transform method of the first class in the MRO after EOFRotator that
        # has a transform method. In this case, it will be HilbertEOF. However,
        # please note that `transform` is not implemented for HilbertEOF, so this
        # line of code will actually raise an error.
        return super(EOFRotator, self)._transform_algorithm(data)



import numpy as np
import pytest
import xarray as xr

from xeofs.single import OPA


@pytest.fixture
def opa_model():
    return OPA(n_modes=3, tau_max=3, n_pca_modes=19)


def test_init():
    """Tests the initialization of the OPA class"""
    opa = OPA(n_modes=3, tau_max=3, n_pca_modes=19, use_coslat=True)

    # Assert preprocessor has been initialized
    assert hasattr(opa, "_params")
    assert hasattr(opa, "preprocessor")


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_fit(dim, mock_data_array, opa_model):
    """Tests the fit method of the OPA class"""

    opa_model.fit(mock_data_array, dim)

    # Assert the required attributes have been set
    assert hasattr(opa_model, "preprocessor")
    assert hasattr(opa_model, "data")


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_decorrelation_time(dim, mock_data_array, opa_model):
    """Tests the decorrelation_time method of the OPA class"""

    opa_model.fit(mock_data_array, dim)

    # Test decorrelation time method
    decorrelation_time = opa_model.decorrelation_time()
    assert isinstance(decorrelation_time, xr.DataArray)

    # decorrelation times are all positive
    assert (decorrelation_time > 0).all()


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components(dim, mock_data_array, opa_model):
    """Tests the components method of the OPA class"""
    opa_model.fit(mock_data_array, dim)

    # Test components method
    components = opa_model.components()
    feature_dims = tuple(set(mock_data_array.dims) - set(dim))
    assert isinstance(components, xr.DataArray), "Components is not a DataArray"
    assert set(components.dims) == set(
        ("mode",) + feature_dims
    ), "Components does not have the right feature dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_filter_patterns(dim, mock_data_array, opa_model):
    """Tests the filter_patterns method of the OPA class"""
    opa_model.fit(mock_data_array, dim)

    # Test components method
    components = opa_model.filter_patterns()
    feature_dims = tuple(set(mock_data_array.dims) - set(dim))
    assert isinstance(components, xr.DataArray), "Components is not a DataArray"
    assert set(components.dims) == set(
        ("mode",) + feature_dims
    ), "Components does not have the right feature dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components_dataset(dim, mock_dataset, opa_model):
    """Tests the components method of the OPA class"""
    opa_model.fit(mock_dataset, dim)

    # Test components method
    components = opa_model.components()
    feature_dims = tuple(set(mock_dataset.dims) - set(dim))
    assert isinstance(components, xr.Dataset), "Components is not a Dataset"
    assert set(components.data_vars) == set(
        mock_dataset.data_vars
    ), "Components does not have the same data variables as the input Dataset"
    assert set(components.dims) == set(
        ("mode",) + feature_dims
    ), "Components does not have the right feature dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_components_dataarray_list(dim, mock_data_array_list, opa_model):
    """Tests the components method of the OPA class"""
    opa_model.fit(mock_data_array_list, dim)

    # Test components method
    components = opa_model.components()
    feature_dims = [tuple(set(data.dims) - set(dim)) for data in mock_data_array_list]
    assert isinstance(components, list), "Components is not a list"
    assert len(components) == len(
        mock_data_array_list
    ), "Components does not have the same length as the input list"
    assert isinstance(
        components[0], xr.DataArray
    ), "Components is not a list of DataArrays"
    for comp, feat_dims in zip(components, feature_dims):
        assert set(comp.dims) == set(
            ("mode",) + feat_dims
        ), "Components does not have the right feature dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores(dim, mock_data_array, opa_model):
    """Tests the scores method of the OPA class"""
    opa_model.fit(mock_data_array, dim)

    # Test scores method
    scores = opa_model.scores()
    assert isinstance(scores, xr.DataArray), "Scores is not a DataArray"
    assert set(scores.dims) == set(
        (dim + ("mode",))
    ), "Scores does not have the right dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores_dataset(dim, mock_dataset, opa_model):
    """Tests the scores method of the OPA class"""
    opa_model.fit(mock_dataset, dim)

    # Test scores method
    scores = opa_model.scores()
    assert isinstance(scores, xr.DataArray)
    assert set(scores.dims) == set(
        (dim + ("mode",))
    ), "Scores does not have the right dimensions"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_scores_dataarray_list(dim, mock_data_array_list, opa_model):
    """Tests the scores method of the OPA class"""
    opa_model.fit(mock_data_array_list, dim)

    # Test scores method
    scores = opa_model.scores()
    assert isinstance(scores, xr.DataArray)
    assert set(scores.dims) == set(
        (dim + ("mode",))
    ), "Scores does not have the right dimensions"


def test_get_params(opa_model):
    """Tests the get_params method of the OPA class"""
    # Test get_params method
    params = opa_model.get_params()
    assert isinstance(params, dict)
    assert params.get("n_modes") == 3
    assert params.get("tau_max") == 3
    assert params.get("n_pca_modes") == 19
    assert params.get("standardize") is False
    assert params.get("use_coslat") is False
    assert params.get("solver") == "auto"


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_transform(dim, mock_data_array, opa_model):
    """Transform is not implemented for OPA"""

    # Create a xarray DataArray with random data
    opa_model.fit(mock_data_array, dim)

    with pytest.raises(NotImplementedError):
        opa_model.transform(mock_data_array)
    # scores = opa_model.scores()

    # # Create a new xarray DataArray with random data
    # new_data = mock_data_array

    # projections = opa_model.transform(new_data)

    # # Check that the projection has the right dimensions
    # assert projections.dims == scores.dims, "Projection has wrong dimensions"  # type: ignore

    # # Check that the projection has the right data type
    # assert isinstance(projections, xr.DataArray), "Projection is not a DataArray"

    # # Check that the projection has the right name
    # assert projections.name == "scores", "Projection has wrong name"

    # # Check that the projection's data is the same as the scores
    # np.testing.assert_allclose(
    #     scores.sel(mode=slice(1, 3)), projections.sel(mode=slice(1, 3)), rtol=1e-3
    # )


@pytest.mark.parametrize(
    "dim",
    [
        (("time",)),
        (("lat", "lon")),
        (("lon", "lat")),
    ],
)
def test_inverse_transform(dim, mock_data_array, opa_model):
    """Test inverse_transform method in EOF class."""

    # fit the EOF model
    opa_model.fit(mock_data_array, dim=dim)
    scores = opa_model.scores()

    with pytest.raises(NotImplementedError):
        opa_model.inverse_transform(scores)

    # # Test with scalar
    # mode = 1
    # reconstructed_data = opa_model.inverse_transform(mode)
    # assert isinstance(reconstructed_data, xr.DataArray)

    # # Test with slice
    # mode = slice(1, 2)
    # reconstructed_data = opa_model.inverse_transform(mode)
    # assert isinstance(reconstructed_data, xr.DataArray)

    # # Test with array of tick labels
    # mode = np.array([1, 3])
    # reconstructed_data = opa_model.inverse_transform(mode)
    # assert isinstance(reconstructed_data, xr.DataArray)

    # # Check that the reconstructed data has the same dimensions as the original data
    # assert set(reconstructed_data.dims) == set(mock_data_array.dims)


# Check mathematical properties
# =============================================================================
@pytest.mark.parametrize(
    "dim, use_coslat",
    [
        (("time",), True),
        (("lat", "lon"), False),
        (("lon", "lat"), False),
    ],
)
def test_U_orthogonal(dim, use_coslat, mock_data_array):
    """U patterns are orthogonal"""
    model = OPA(
        n_modes=3, tau_max=3, n_pca_modes=19, standardize=True, use_coslat=use_coslat
    )
    model.fit(mock_data_array, dim=dim)
    U = model._U.values
    assert np.allclose(
        U.T @ U, np.eye(U.shape[1]), atol=1e-5
    ), "U patterns are not orthogonal"


# IGNORE THIS TEST: Current implementation yields C0 in PCA space and FP in original space
# therefore the test fails
# @pytest.mark.parametrize(
#     "dim, use_coslat",
#     [
#         (("time",), True),
#         (("lat", "lon"), False),
#         (("lon", "lat"), False),
#     ],
# )
# def test_filter_patterns_biorthogonal(dim, use_coslat, mock_data_array):
#     """Filter patterns are biorthogonal"""
#     n_pca_modes = 20
#     model = OPA(
#         n_modes=3,
#         tau_max=3,
#         n_pca_modes=n_pca_modes,
#         standardize=True,
#         use_coslat=use_coslat,
#     )
#     model.fit(mock_data_array, dim=dim)
#     FP = model.data.filter_patterns.values
#     C0 = model._C0.values  # zero-lag covariance matrix
#     assert C0.shape == (n_pca_modes, n_pca_modes)
#     check = FP.T @ C0 @ FP
#     assert np.allclose(
#         check, np.eye(check.shape[1]), atol=1e-5
#     ), "Filter patterns are are not biorthogonal"


@pytest.mark.parametrize(
    "dim, use_coslat",
    [
        (("time",), True),
        (("lat", "lon"), False),
        (("lon", "lat"), False),
    ],
)
def test_scores_uncorrelated(dim, use_coslat, mock_data_array):
    """Scores are uncorrelated"""
    n_pca_modes = 20
    model = OPA(
        n_modes=3,
        tau_max=3,
        n_pca_modes=n_pca_modes,
        standardize=True,
        use_coslat=use_coslat,
    )
    model.fit(mock_data_array, dim=dim)
    scores = model.data["scores"].values
    check = scores.T @ scores / (scores.shape[0] - 1)
    assert np.allclose(
        check, np.eye(check.shape[1]), atol=1e-5
    ), "Scores are not uncorrelated"
